# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os

from tacker import context
from tacker.sol_refactored.common import vnfd_utils
from tacker.sol_refactored.infra_drivers.openstack import openstack
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields
from tacker.tests import base


SAMPLE_VNFD_ID = "b1bb0ce7-ebca-4fa7-95ed-4840d7000000"
SAMPLE_FLAVOUR_ID = "simple"

# instantiateVnfRequest example
_vim_connection_info_example = {
    "vimId": "vim_id_1",
    "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
    # "interfaceInfo": omitted
    # "accessInfo": omitted
}

_instantiate_req_example = {
    "flavourId": SAMPLE_FLAVOUR_ID,
    "extVirtualLinks": [
        {
            "id": "id_ext_vl_1",
            "resourceId": "res_id_ext_vl_1",
            "extCps": [
                {
                    "cpdId": "VDU1_CP1",
                    "cpConfig": {
                        "VDU1_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP1",
                    "cpConfig": {
                        "VDU2_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "10.10.0.102"
                                                ]
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {
            "id": "id_ext_vl_2",
            "resourceId": "res_id_id_ext_vl_2",
            "extCps": [
                {
                    "cpdId": "VDU1_CP2",
                    "cpConfig": {
                        "VDU1_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1,
                                                "subnetId": "res_id_subnet_1"
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP2",
                    "cpConfig": {
                        "VDU2_CP2_1": {
                            "linkPortId": "link_port_id_VDU2_CP2"
                        }
                    }
                }
            ],
            "extLinkPorts": [
                {
                    "id": "link_port_id_VDU2_CP2",
                    "resourceHandle": {
                        "resourceId": "res_id_VDU2_CP2"
                    }
                }
            ]
        }
    ],
    "extManagedVirtualLinks": [
        {
            "id": "id_ext_mgd_1",
            "vnfVirtualLinkDescId": "internalVL1",
            "resourceId": "res_id_internalVL1"
        }
    ],
    "vimConnectionInfo": {
        "vim1": _vim_connection_info_example
    }
}

# ChangeExtVnfConnectivityRequest example
_change_ext_conn_req_example = {
    "extVirtualLinks": [
        {
            "id": "id_ext_vl_3",
            "resourceId": "res_id_ext_vl_3",
            "extCps": [
                {
                    "cpdId": "VDU2_CP1",
                    "cpConfig": {
                        "VDU2_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "20.10.0.102"
                                                ]
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {
            "id": "id_ext_vl_4",
            "resourceId": "res_id_id_ext_vl_4",
            "extCps": [
                {
                    "cpdId": "VDU1_CP2",
                    "cpConfig": {
                        "VDU1_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1,
                                                "subnetId": "res_id_subnet_4"
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP2",
                    "cpConfig": {
                        "VDU2_CP2_1": {
                            "linkPortId": "link_port_id_VDU2_CP2_modified"
                        }
                    }
                }
            ],
            "extLinkPorts": [
                {
                    "id": "link_port_id_VDU2_CP2_modified",
                    "resourceHandle": {
                        "resourceId": "res_id_VDU2_CP2_modified"
                    }
                }
            ]
        }
    ]
}

# heat resources examples
# NOTE:
# - following attributes which are not related to tests are omitted.
#   updated_time, logical_resource_id, resource_status, resource_status_reason
#   and "rel: self" in links.
# - some identifiers are modified to make check easy.
# - stack_id is based on real example.
_url = "http://127.0.0.1/heat-api/v1/57bcfdcbccbc4b85a9a4c94690a1164f/stacks/"
_stack_id = ("vnf-768c24d2-2ea6-4225-b1c7-79e42abfbde6/"
             "1fa212ca-e904-4109-9a96-1900d35d2a5b")
_href = "".join((_url, _stack_id))

_stack_id_VDU1_scale = (
    "vnf-768c24d2-2ea6-4225-b1c7-79e42abfbde6-VDU1_scale_group-dv4kv7qtcwhw/"
    "53ee92b6-8193-4df5-90f7-2738e61fba2c")
_href_VDU1_scale = "".join((_url, _stack_id_VDU1_scale))

_stack_id_VDU1_1 = (
    "vnf-768c24d2-2ea6-4225-b1c7-79e42abfbde6-VDU1_scale_group-dv4kv7qtcwhw-"
    "bemybz4ugeso-mrajuhqw7ath/ea59d312-bab3-4ef2-897c-88b5cee117de")
_href_VDU1_1 = "".join((_url, _stack_id_VDU1_1))

_stack_id_VDU1_2 = (
    "vnf-768c24d2-2ea6-4225-b1c7-79e42abfbde6-VDU1_scale_group-dv4kv7qtcwhw-"
    "myet4efobvvp-aptv6apap2h5/dd94d2ae-a02b-4fab-a492-514c422299ec")
_href_VDU1_2 = "".join((_url, _stack_id_VDU1_2))

_heat_reses_example_base = [
    {
        "creation_time": "2021-12-10T00:40:46Z",
        "resource_name": "VDU2",
        "physical_resource_id": "res_id_VDU2",
        "resource_type": "OS::Nova::Server",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": []
    },
    {
        "creation_time": "2021-12-10T00:40:47Z",
        "resource_name": "VDU2_CP5",
        "physical_resource_id": "res_id_VDU2_CP5",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:49Z",
        "resource_name": "internalVL3_subnet",
        "physical_resource_id": "06f68f37-d37c-4310-8756-3884d9b8cb4b",
        "resource_type": "OS::Neutron::Subnet",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2",
            "VDU2_CP5"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:51Z",
        "resource_name": "VDU2_CP3",
        "physical_resource_id": "res_id_VDU2_CP3",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:53Z",
        "resource_name": "VDU2_CP4",
        "physical_resource_id": "res_id_VDU2_CP4",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:53Z",
        "resource_name": "VDU1_scale_out",
        "physical_resource_id": "234c78aefbaa4770ba5bbd8c7f624584",
        "resource_type": "OS::Heat::ScalingPolicy",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": []
    },
    {
        "creation_time": "2021-12-10T00:40:53Z",
        "resource_name": "VDU1_scale_in",
        "physical_resource_id": "7cb738b392a64936a14f92bb79123a42",
        "resource_type": "OS::Heat::ScalingPolicy",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": []
    },
    {
        "creation_time": "2021-12-10T00:40:53Z",
        "resource_name": "VDU1_scale_group",
        "physical_resource_id": "53ee92b6-8193-4df5-90f7-2738e61fba2c",
        "resource_type": "OS::Heat::AutoScalingGroup",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            },
            {
                "href": _href_VDU1_scale,
                "rel": "nested"
            }
        ],
        "required_by": [
            "VDU1_scale_out",
            "VDU1_scale_in"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:55Z",
        "resource_name": "internalVL3",
        "physical_resource_id": "res_id_internalVL3",
        "resource_type": "OS::Neutron::Net",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "internalVL3_subnet",
            "VDU2_CP5",
            "VDU1_scale_group"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:56Z",
        "resource_name": "internalVL2_subnet",
        "physical_resource_id": "848bc969-cc5c-47b2-94de-469556b993fb",
        "resource_type": "OS::Neutron::Subnet",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2",
            "VDU2_CP4"
        ]
    },
    {
        "creation_time": "2021-12-10T00:40:58Z",
        "resource_name": "internalVL2",
        "physical_resource_id": "res_id_internalVL2",
        "resource_type": "OS::Neutron::Net",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2_CP4",
            "internalVL2_subnet",
            "VDU1_scale_group"
        ]
    },
    {
        "creation_time": "2021-12-10T00:41:35Z",
        "resource_name": "bemybz4ugeso",
        "physical_resource_id": "ea59d312-bab3-4ef2-897c-88b5cee117de",
        "resource_type": "VDU1.yaml",
        "links": [
            {
                "href": _href_VDU1_scale,
                "rel": "stack"
            },
            {
                "href": _href_VDU1_1,
                "rel": "nested"
            }
        ],
        "required_by": [],
        "parent_resource": "VDU1_scale_group"
    },
    {
        "creation_time": "2021-12-10T01:03:37Z",
        "resource_name": "myet4efobvvp",
        "physical_resource_id": "dd94d2ae-a02b-4fab-a492-514c422299ec",
        "resource_type": "VDU1.yaml",
        "links": [
            {
                "href": _href_VDU1_scale,
                "rel": "stack"
            },
            {
                "href": _href_VDU1_2,
                "rel": "nested"
            }
        ],
        "required_by": [],
        "parent_resource": "VDU1_scale_group"
    },
    {
        "creation_time": "2021-12-10T00:41:43Z",
        "resource_name": "VDU1",
        "physical_resource_id": "res_id_VDU1_1",
        "resource_type": "OS::Nova::Server",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "VirtualStorage",
        "physical_resource_id": "res_id_VirtualStorage_1",
        "resource_type": "OS::Cinder::Volume",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "multi",
        "physical_resource_id": "0690bb1b-36d0-4684-851b-9c1b13a9a5de",
        "resource_type": "OS::Cinder::VolumeType",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VirtualStorage"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "VDU1_CP1",
        "physical_resource_id": "res_id_VDU1_CP1_1",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "multi",
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "VDU1_CP4",
        "physical_resource_id": "res_id_VDU1_CP4_1",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "VDU1_CP5",
        "physical_resource_id": "res_id_VDU1_CP5_1",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T00:41:46Z",
        "resource_name": "VDU1_CP3",
        "physical_resource_id": "res_id_VDU1_CP3_1",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T01:03:49Z",
        "resource_name": "VDU1",
        "physical_resource_id": "res_id_VDU1_2",
        "resource_type": "OS::Nova::Server",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [],
        "parent_resource": "myet4efobvvp"
    },
    {
        "creation_time": "2021-12-10T01:03:51Z",
        "resource_name": "VirtualStorage",
        "physical_resource_id": "res_id_VirtualStorage_2",
        "resource_type": "OS::Cinder::Volume",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    },
    {
        "creation_time": "2021-12-10T01:03:53Z",
        "resource_name": "multi",
        "physical_resource_id": "03433e3a-a8c4-40b3-8802-6b114922feff",
        "resource_type": "OS::Cinder::VolumeType",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VirtualStorage"
        ],
        "parent_resource": "myet4efobvvp"
    },
    {
        "creation_time": "2021-12-10T01:03:53Z",
        "resource_name": "VDU1_CP1",
        "physical_resource_id": "res_id_VDU1_CP1_2",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "multi",
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    },
    {
        "creation_time": "2021-12-10T01:03:53Z",
        "resource_name": "VDU1_CP4",
        "physical_resource_id": "res_id_VDU1_CP4_2",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    },
    {
        "creation_time": "2021-12-10T01:03:55Z",
        "resource_name": "VDU1_CP5",
        "physical_resource_id": "res_id_VDU1_CP5_2",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    },
    {
        "creation_time": "2021-12-10T01:03:57Z",
        "resource_name": "VDU1_CP3",
        "physical_resource_id": "res_id_VDU1_CP3_2",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    }
]

_heat_reses_example_cps_before = [
    {
        "creation_time": "2021-12-10T00:40:46Z",
        "resource_name": "VDU2_CP1",
        "physical_resource_id": "res_id_VDU2_CP1",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2"
        ]
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "VDU1_CP2",
        "physical_resource_id": "res_id_VDU1_CP2_1",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T01:03:53Z",
        "resource_name": "VDU1_CP2",
        "physical_resource_id": "res_id_VDU1_CP2_2",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    }
]

_heat_reses_example_cps_after = [
    {
        "creation_time": "2021-12-10T00:40:46Z",
        "resource_name": "VDU2_CP1",
        "physical_resource_id": "res_id_VDU2_CP1_modified",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU2"
        ]
    },
    {
        "creation_time": "2021-12-10T00:41:45Z",
        "resource_name": "VDU1_CP2",
        "physical_resource_id": "res_id_VDU1_CP2_1_modified",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_1,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "bemybz4ugeso"
    },
    {
        "creation_time": "2021-12-10T01:03:53Z",
        "resource_name": "VDU1_CP2",
        "physical_resource_id": "res_id_VDU1_CP2_2_modified",
        "resource_type": "OS::Neutron::Port",
        "links": [
            {
                "href": _href_VDU1_2,
                "rel": "stack"
            }
        ],
        "required_by": [
            "VDU1"
        ],
        "parent_resource": "myet4efobvvp"
    }
]

# heat resources example for other than change_ext_conn
_heat_reses_example = (
    _heat_reses_example_base + _heat_reses_example_cps_before)

# heat resources example after executing change_ext_conn
_heat_reses_example_change_ext_conn = (
    _heat_reses_example_base + _heat_reses_example_cps_after)

# expected results (other than change_ext_conn)
_expected_inst_info = {
    "flavourId": "simple",
    "vnfState": "STARTED",
    "extCpInfo": [
        {
            'id': 'cp-req-link_port_id_VDU2_CP2',
            'cpdId': 'VDU2_CP2',
            'cpConfigId': 'VDU2_CP2_1',
            'extLinkPortId': 'req-link_port_id_VDU2_CP2',
        },
        {
            "id": "cp-res_id_VDU1_CP1_1",
            "cpdId": "VDU1_CP1",
            "cpConfigId": "VDU1_CP1_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP1_1",
            "associatedVnfcCpId": "VDU1_CP1-res_id_VDU1_1"
        },
        {
            "id": "cp-res_id_VDU1_CP1_2",
            "cpdId": "VDU1_CP1",
            "cpConfigId": "VDU1_CP1_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP1_2",
            "associatedVnfcCpId": "VDU1_CP1-res_id_VDU1_2"
        },
        {
            "id": "cp-res_id_VDU1_CP2_1",
            "cpdId": "VDU1_CP2",
            "cpConfigId": "VDU1_CP2_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True,
                                "subnetId": "res_id_subnet_1"
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP2_1",
            "associatedVnfcCpId": "VDU1_CP2-res_id_VDU1_1"
        },
        {
            "id": "cp-res_id_VDU1_CP2_2",
            "cpdId": "VDU1_CP2",
            "cpConfigId": "VDU1_CP2_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True,
                                "subnetId": "res_id_subnet_1"
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP2_2",
            "associatedVnfcCpId": "VDU1_CP2-res_id_VDU1_2"
        },
        {
            "id": "cp-res_id_VDU2_CP1",
            "cpdId": "VDU2_CP1",
            "cpConfigId": "VDU2_CP1_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "addresses": [
                                    "10.10.0.102"
                                ]
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU2_CP1",
            "associatedVnfcCpId": "VDU2_CP1-res_id_VDU2"
        }
    ],
    "extVirtualLinkInfo": [
        {
            "id": "id_ext_vl_1",
            "resourceHandle": {
                "resourceId": "res_id_ext_vl_1"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU1_CP1_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP1_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "res_id_VDU1_CP1_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP1_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_2"
                },
                {
                    "id": "res_id_VDU2_CP1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU2_CP1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP1"
                }
            ],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP1",
                    "cpConfig": {
                        "VDU1_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP1",
                    "cpConfig": {
                        "VDU2_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "10.10.0.102"
                                                ]
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {
            "id": "id_ext_vl_2",
            "resourceHandle": {
                "resourceId": "res_id_id_ext_vl_2"
            },
            "extLinkPorts": [
                {
                    "id": "req-link_port_id_VDU2_CP2",
                    "resourceHandle": {
                        "resourceId": "res_id_VDU2_CP2",
                    },
                    "cpInstanceId": "cp-req-link_port_id_VDU2_CP2"
                },
                {
                    "id": "res_id_VDU1_CP2_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP2_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_1"
                },
                {
                    "id": "res_id_VDU1_CP2_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP2_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_2"
                }
            ],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP2",
                    "cpConfig": {
                        "VDU1_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1,
                                                "subnetId": "res_id_subnet_1"
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP2",
                    "cpConfig": {
                        "VDU2_CP2_1": {
                            "linkPortId": "link_port_id_VDU2_CP2"
                        }
                    }
                }
            ]
        }
    ],
    "extManagedVirtualLinkInfo": [
        {
            "id": "id_ext_mgd_1",
            "vnfVirtualLinkDescId": "internalVL1",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU1_CP3_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP3_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP3_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP3_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU2_CP3",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU2_CP3",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP3-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "vnfcResourceInfo": [
        {
            "id": "res_id_VDU1_2",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_2"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_2",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_2"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_2",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_2"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_2",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_2"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_2",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_2"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_2",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_2"
                }
            ],
            "metadata": {
                "creation_time": "2021-12-10T01:03:49Z",
                "parent_stack_id": _stack_id_VDU1_scale,
                "parent_resource_name": "myet4efobvvp",
                "stack_id": _stack_id_VDU1_2
            }
        },
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_1"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_1",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_1",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_1"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_1",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_1"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_1",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_1"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_1",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_1"
                }
            ],
            "metadata": {
                "creation_time": "2021-12-10T00:41:43Z",
                "parent_stack_id": _stack_id_VDU1_scale,
                "parent_resource_name": "bemybz4ugeso",
                "stack_id": _stack_id_VDU1_1
            }
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "computeResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "vnfcCpInfo": [
                {
                    "id": "VDU2_CP1-res_id_VDU2",
                    "cpdId": "VDU2_CP1",
                    "vnfExtCpId": "cp-res_id_VDU2_CP1"
                },
                {
                    "id": "VDU2_CP2-res_id_VDU2",
                    "cpdId": "VDU2_CP2",
                    # "vnfExtCpId" does not exist since it is specified by
                    # linkPortIds.
                },
                {
                    "id": "VDU2_CP3-res_id_VDU2",
                    "cpdId": "VDU2_CP3",
                    "vnfLinkPortId": "res_id_VDU2_CP3"
                },
                {
                    "id": "VDU2_CP4-res_id_VDU2",
                    "cpdId": "VDU2_CP4",
                    "vnfLinkPortId": "res_id_VDU2_CP4"
                },
                {
                    "id": "VDU2_CP5-res_id_VDU2",
                    "cpdId": "VDU2_CP5",
                    "vnfLinkPortId": "res_id_VDU2_CP5"
                }
            ],
            "metadata": {
                "creation_time": "2021-12-10T00:40:46Z",
                "stack_id": _stack_id
            }
        }
    ],
    "vnfVirtualLinkResourceInfo": [
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "networkResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU1_CP4_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP4_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP4-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP4_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP4_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP4-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU2_CP4",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU2_CP4",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP4-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        },
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "networkResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU1_CP5_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP5_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP5-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP5_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP5_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP5-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU2_CP5",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU2_CP5",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP5-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "virtualStorageResourceInfo": [
        {
            "id": "res_id_VirtualStorage_1",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            },
            "metadata": {"stack_id": _stack_id_VDU1_1}
        },
        {
            "id": "res_id_VirtualStorage_2",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VirtualStorage_2",
                "vimLevelResourceType": "OS::Cinder::Volume"
            },
            "metadata": {"stack_id": _stack_id_VDU1_2}
        }
    ],
    "vnfcInfo": [
        {
            "id": "VDU1-res_id_VDU1_2",
            "vduId": "VDU1",
            "vnfcResourceInfoId": "res_id_VDU1_2",
            "vnfcState": "STARTED"
        },
        {
            "id": "VDU1-res_id_VDU1_1",
            "vduId": "VDU1",
            "vnfcResourceInfoId": "res_id_VDU1_1",
            "vnfcState": "STARTED"
        },
        {
            "id": "VDU2-res_id_VDU2",
            "vduId": "VDU2",
            "vnfcResourceInfoId": "res_id_VDU2",
            "vnfcState": "STARTED"
        }
    ]
}

_expected_inst_info_vnfc_updated = copy.copy(_expected_inst_info)
_expected_inst_info_vnfc_updated["vnfcInfo"] = [
    {
        "id": "VDU1-res_id_VDU1_2",
        "vduId": "VDU1",
        "vnfcResourceInfoId": "res_id_VDU1_2",
        "vnfcState": "STARTED"
    },
    {
        "id": "VDU1-res_id_VDU1_1",
        "vduId": "VDU1",
        "vnfcResourceInfoId": "res_id_VDU1_1",
        "vnfcState": "STARTED",
        "vnfcConfigurableProperties": {"prop1": "value1"}
    },
    {
        "id": "VDU2-res_id_VDU2",
        "vduId": "VDU2",
        "vnfcResourceInfoId": "res_id_VDU2",
        "vnfcState": "STARTED"
    }
]

# expected results for change_ext_conn
_expected_inst_info_change_ext_conn = {
    "flavourId": "simple",
    "vnfState": "STARTED",
    "extCpInfo": [
        {
            'id': 'cp-req-link_port_id_VDU2_CP2_modified',
            'cpdId': 'VDU2_CP2',
            'cpConfigId': 'VDU2_CP2_1',
            'extLinkPortId': 'req-link_port_id_VDU2_CP2_modified',
        },
        {
            "id": "cp-res_id_VDU1_CP1_1",
            "cpdId": "VDU1_CP1",
            "cpConfigId": "VDU1_CP1_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP1_1",
            "associatedVnfcCpId": "VDU1_CP1-res_id_VDU1_1"
        },
        {
            "id": "cp-res_id_VDU1_CP1_2",
            "cpdId": "VDU1_CP1",
            "cpConfigId": "VDU1_CP1_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP1_2",
            "associatedVnfcCpId": "VDU1_CP1-res_id_VDU1_2"
        },
        {
            "id": "cp-res_id_VDU1_CP2_1_modified",
            "cpdId": "VDU1_CP2",
            "cpConfigId": "VDU1_CP2_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True,
                                "subnetId": "res_id_subnet_4"
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP2_1_modified",
            "associatedVnfcCpId": "VDU1_CP2-res_id_VDU1_1"
        },
        {
            "id": "cp-res_id_VDU1_CP2_2_modified",
            "cpdId": "VDU1_CP2",
            "cpConfigId": "VDU1_CP2_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "isDynamic": True,
                                "subnetId": "res_id_subnet_4"
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU1_CP2_2_modified",
            "associatedVnfcCpId": "VDU1_CP2-res_id_VDU1_2"
        },
        {
            "id": "cp-res_id_VDU2_CP1_modified",
            "cpdId": "VDU2_CP1",
            "cpConfigId": "VDU2_CP1_1",
            "cpProtocolInfo": [
                {
                    "layerProtocol": "IP_OVER_ETHERNET",
                    "ipOverEthernet": {
                        "ipAddresses": [
                            {
                                "type": "IPV4",
                                "addresses": [
                                    "20.10.0.102"
                                ]
                            }
                        ]
                    }
                }
            ],
            "extLinkPortId": "res_id_VDU2_CP1_modified",
            "associatedVnfcCpId": "VDU2_CP1-res_id_VDU2"
        }
    ],
    "extVirtualLinkInfo": [
        {
            "id": "id_ext_vl_1",
            "resourceHandle": {
                "resourceId": "res_id_ext_vl_1"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU1_CP1_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP1_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "res_id_VDU1_CP1_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP1_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_2"
                }
            ],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP1",
                    "cpConfig": {
                        "VDU1_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {
            "id": "id_ext_vl_3",
            "resourceHandle": {
                "resourceId": "res_id_ext_vl_3"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP1_modified",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU2_CP1_modified",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP1_modified"
                }
            ],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU2_CP1",
                    "cpConfig": {
                        "VDU2_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "20.10.0.102"
                                                ]
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {
            "id": "id_ext_vl_4",
            "resourceHandle": {
                "resourceId": "res_id_id_ext_vl_4"
            },
            "extLinkPorts": [
                {
                    "id": "req-link_port_id_VDU2_CP2_modified",
                    "resourceHandle": {
                        "resourceId": "res_id_VDU2_CP2_modified",
                    },
                    "cpInstanceId": "cp-req-link_port_id_VDU2_CP2_modified"
                },
                {
                    "id": "res_id_VDU1_CP2_1_modified",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP2_1_modified",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_1_modified"
                },
                {
                    "id": "res_id_VDU1_CP2_2_modified",
                    "resourceHandle": {
                        "vimConnectionId": "vim_id_1",
                        "resourceId": "res_id_VDU1_CP2_2_modified",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_2_modified"
                }
            ],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP2",
                    "cpConfig": {
                        "VDU1_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1,
                                                "subnetId": "res_id_subnet_4"
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP2",
                    "cpConfig": {
                        "VDU2_CP2_1": {
                            "linkPortId": "link_port_id_VDU2_CP2_modified"
                        }
                    }
                }
            ]
        }
    ],
    "vnfcResourceInfo": [
        {
            "id": "res_id_VDU1_2",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_2"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_2",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_2"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_2",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_2_modified"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_2",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_2"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_2",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_2"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_2",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_2"
                }
            ],
            "metadata": {
                "creation_time": "2021-12-10T01:03:49Z",
                "parent_stack_id": _stack_id_VDU1_scale,
                "parent_resource_name": "myet4efobvvp",
                "stack_id": _stack_id_VDU1_2
            }
        },
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_1"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_1",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_1",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_1_modified"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_1",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_1"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_1",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_1"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_1",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_1"
                }
            ],
            "metadata": {
                "creation_time": "2021-12-10T00:41:43Z",
                "parent_stack_id": _stack_id_VDU1_scale,
                "parent_resource_name": "bemybz4ugeso",
                "stack_id": _stack_id_VDU1_1
            }
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "computeResource": {
                "vimConnectionId": "vim_id_1",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "vnfcCpInfo": [
                {
                    "id": "VDU2_CP1-res_id_VDU2",
                    "cpdId": "VDU2_CP1",
                    "vnfExtCpId": "cp-res_id_VDU2_CP1_modified"
                },
                {
                    "id": "VDU2_CP2-res_id_VDU2",
                    "cpdId": "VDU2_CP2",
                    # "vnfExtCpId" does not exist since it is specified by
                    # linkPortIds.
                },
                {
                    "id": "VDU2_CP3-res_id_VDU2",
                    "cpdId": "VDU2_CP3",
                    "vnfLinkPortId": "res_id_VDU2_CP3"
                },
                {
                    "id": "VDU2_CP4-res_id_VDU2",
                    "cpdId": "VDU2_CP4",
                    "vnfLinkPortId": "res_id_VDU2_CP4"
                },
                {
                    "id": "VDU2_CP5-res_id_VDU2",
                    "cpdId": "VDU2_CP5",
                    "vnfLinkPortId": "res_id_VDU2_CP5"
                }
            ],
            "metadata": {
                "creation_time": "2021-12-10T00:40:46Z",
                "stack_id": _stack_id
            }
        }
    ],
    # other members are not changed from _expected_inst_info
    "extManagedVirtualLinkInfo":
        _expected_inst_info["extManagedVirtualLinkInfo"],
    "vnfVirtualLinkResourceInfo":
        _expected_inst_info["vnfVirtualLinkResourceInfo"],
    "virtualStorageResourceInfo":
        _expected_inst_info["virtualStorageResourceInfo"],
    "vnfcInfo": _expected_inst_info["vnfcInfo"]
}


class TestOpenstack(base.BaseTestCase):

    def setUp(self):
        super(TestOpenstack, self).setUp()
        objects.register_all()
        self.driver = openstack.Openstack()
        self.context = context.get_admin_context()

        cur_dir = os.path.dirname(__file__)
        sample_dir = os.path.join(cur_dir, "../..", "samples")

        self.vnfd_1 = vnfd_utils.Vnfd(SAMPLE_VNFD_ID)
        self.vnfd_1.init_from_csar_dir(os.path.join(sample_dir, "sample1"))

    def _check_inst_info(self, expected, result):
        # sort lists before compare with an expected result since
        # order of list items is unpredictable.
        # note that an expected_result is already sorted.
        def _get_key(obj):
            return obj['id']

        if "extCpInfo" in expected:
            self.assertIn("extCpInfo", result)
            result["extCpInfo"].sort(key=_get_key)
            # assume len(cpProtocolInfo) == 1
            self.assertEqual(expected["extCpInfo"], result["extCpInfo"])

        if "extVirtualLinkInfo" in expected:
            self.assertIn("extVirtualLinkInfo", result)
            result["extVirtualLinkInfo"].sort(key=_get_key)
            for ext_vl in result["extVirtualLinkInfo"]:
                if "extLinkPorts" in ext_vl:
                    ext_vl["extLinkPorts"].sort(key=_get_key)
                # order of currentVnfExtCpData is same as order of
                # instantiateVnfRequest
            self.assertEqual(expected["extVirtualLinkInfo"],
                result["extVirtualLinkInfo"])

        if "extManagedVirtualLinkInfo" in expected:
            self.assertIn("extManagedVirtualLinkInfo", result)
            for ext_mgd in result["extManagedVirtualLinkInfo"]:
                if "vnfLinkPorts" in ext_mgd:
                    ext_mgd["vnfLinkPorts"].sort(key=_get_key)
            self.assertEqual(expected["extManagedVirtualLinkInfo"],
                result["extManagedVirtualLinkInfo"])

        # vnfcResourceInfo is sorted by creation_time (reverse)
        if "vnfcResourceInfo" in expected:
            self.assertIn("vnfcResourceInfo", result)
            for vnfc in result["vnfcResourceInfo"]:
                if "storageResourceIds" in vnfc:
                    vnfc["storageResourceIds"].sort()
                if "vnfcCpInfo" in vnfc:
                    vnfc["vnfcCpInfo"].sort(key=_get_key)
            self.assertEqual(expected["vnfcResourceInfo"],
                result["vnfcResourceInfo"])

        if "vnfVirtualLinkResourceInfo" in expected:
            self.assertIn("vnfVirtualLinkResourceInfo", result)
            result["vnfVirtualLinkResourceInfo"].sort(key=_get_key)
            for vl_info in result["vnfVirtualLinkResourceInfo"]:
                if "vnfLinkPorts" in vl_info:
                    vl_info["vnfLinkPorts"].sort(key=_get_key)
            self.assertEqual(expected["vnfVirtualLinkResourceInfo"],
                result["vnfVirtualLinkResourceInfo"])

        if "virtualStorageResourceInfo" in expected:
            self.assertIn("virtualStorageResourceInfo", result)
            result["virtualStorageResourceInfo"].sort(key=_get_key)
            self.assertEqual(expected["virtualStorageResourceInfo"],
                result["virtualStorageResourceInfo"])

        # order of vnfcInfo is same as vnfcResourceInfo
        if "vnfcInfo" in expected:
            self.assertIn("vnfcInfo", result)
            self.assertEqual(expected["vnfcInfo"], result["vnfcInfo"])

    def test_make_instantiated_vnf_info_new(self):
        # prepare
        req = objects.InstantiateVnfRequest.from_dict(
            _instantiate_req_example)
        inst = objects.VnfInstanceV2(
            vimConnectionInfo=req.vimConnectionInfo
        )
        grant_req = objects.GrantRequestV1(
            operation=fields.LcmOperationType.INSTANTIATE
        )
        grant = objects.GrantV1()

        # execute make_instantiated_vnf_info
        self.driver._make_instantiated_vnf_info(req, inst, grant_req, grant,
            self.vnfd_1, _heat_reses_example)

        # check
        result = inst.to_dict()["instantiatedVnfInfo"]
        self._check_inst_info(_expected_inst_info, result)

    def test_make_instantiated_vnf_info_update(self):
        # prepare
        req = None  # not used
        inst_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _expected_inst_info_vnfc_updated)
        vim_info = {
            "vim1": objects.VimConnectionInfo.from_dict(
                _vim_connection_info_example)
        }
        inst = objects.VnfInstanceV2(
            instantiatedVnfInfo=inst_info,
            vimConnectionInfo=vim_info
        )
        grant_req = objects.GrantRequestV1(
            operation=fields.LcmOperationType.SCALE
        )
        grant = objects.GrantV1()

        # execute make_instantiated_vnf_info
        self.driver._make_instantiated_vnf_info(req, inst, grant_req, grant,
            self.vnfd_1, _heat_reses_example)

        # check
        result = inst.to_dict()["instantiatedVnfInfo"]
        self._check_inst_info(_expected_inst_info_vnfc_updated, result)

    def test_make_instantiated_vnf_info_change_ext_conn(self):
        # prepare
        req = objects.ChangeExtVnfConnectivityRequest.from_dict(
            _change_ext_conn_req_example)
        inst_info = objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
            _expected_inst_info)
        vim_info = {
            "vim1": objects.VimConnectionInfo.from_dict(
                _vim_connection_info_example)
        }
        inst = objects.VnfInstanceV2(
            instantiatedVnfInfo=inst_info,
            vimConnectionInfo=vim_info
        )
        grant_req = objects.GrantRequestV1(
            operation=fields.LcmOperationType.CHANGE_EXT_CONN
        )
        grant = objects.GrantV1()

        # execute make_instantiated_vnf_info
        self.driver._make_instantiated_vnf_info(req, inst, grant_req, grant,
            self.vnfd_1, _heat_reses_example_change_ext_conn)

        # check
        result = inst.to_dict()["instantiatedVnfInfo"]
        self._check_inst_info(_expected_inst_info_change_ext_conn, result)



# Copyright 2013 VMware, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for extension extended attribute
"""

import webob.exc as webexc

import tacker
from tacker.api import extensions
from tacker.api.v1 import attributes
from tacker.common import config
from tacker import manager
from tacker.tests import base
from tacker.tests.unit.api.v2 import test_api_v2
from tacker.tests.unit.extensions import extendedattribute as extattr
from tacker.tests.unit import testlib_api
from tacker import wsgi

_uuid = test_api_v2._uuid
_get_path = test_api_v2._get_path
extensions_path = ':'.join(tacker.tests.unit.extensions.__path__)


class ExtensionExtendedAttributeTestCase(base.BaseTestCase):
    def setUp(self):
        super(ExtensionExtendedAttributeTestCase, self).setUp()
        self.skipTest("Not ready yet")
        plugin = (
            "tacker.tests.unit.test_extension_extended_attribute."
            "ExtensionExtendedAttributeTestPlugin"
        )

        # point config file to: tacker/tests/etc/tacker.conf.test
        self.config_parse()

        self.setup_coreplugin(plugin)

        ext_mgr = extensions.ExtensionManager(extensions_path)
        ext_mgr.extend_resources("1.0", {})
        extensions.ExtensionManager._instance = ext_mgr

        app = config.load_paste_app('extensions_test_app')
        self._api = extensions.ExtensionMiddleware(app, ext_mgr=ext_mgr)

        self._tenant_id = "8c70909f-b081-452d-872b-df48e6c355d1"
        # Save the global RESOURCE_ATTRIBUTE_MAP
        self.saved_attr_map = {}
        for resource, attrs in (attributes.RESOURCE_ATTRIBUTE_MAP).items():
            self.saved_attr_map[resource] = attrs.copy()
        # Add the resources to the global attribute map
        # This is done here as the setup process won't
        # initialize the main API router which extends
        # the global attribute map
        attributes.RESOURCE_ATTRIBUTE_MAP.update(
            extattr.EXTENDED_ATTRIBUTES_2_0)
        self.agentscheduler_dbMinxin = manager.TackerManager.get_plugin()
        self.addCleanup(self.restore_attribute_map)

    def restore_attribute_map(self):
        # Restore the original RESOURCE_ATTRIBUTE_MAP
        attributes.RESOURCE_ATTRIBUTE_MAP = self.saved_attr_map

    def _do_request(self, method, path, data=None, params=None, action=None):
        content_type = 'application/json'
        body = None
        if data is not None:  # empty dict is valid
            body = wsgi.Serializer().serialize(data, content_type)

        req = testlib_api.create_request(
            path, body, content_type,
            method, query_string=params)
        res = req.get_response(self._api)
        if res.status_code >= 400:
            raise webexc.HTTPClientError(detail=res.body, code=res.status_code)
        if res.status_code != webexc.HTTPNoContent.code:
            return res.json

    def _ext_test_resource_create(self, attr=None):
        data = {
            "ext_test_resource": {
                "tenant_id": self._tenant_id,
                "name": "test",
                extattr.EXTENDED_ATTRIBUTE: attr
            }
        }

        res = self._do_request('POST', _get_path('ext_test_resources'), data)
        return res['ext_test_resource']

    def test_ext_test_resource_create(self):
        ext_test_resource = self._ext_test_resource_create()
        attr = _uuid()
        ext_test_resource = self._ext_test_resource_create(attr)
        self.assertEqual(attr, ext_test_resource[extattr.EXTENDED_ATTRIBUTE])

    def test_ext_test_resource_get(self):
        attr = _uuid()
        obj = self._ext_test_resource_create(attr)
        obj_id = obj['id']
        res = self._do_request('GET', _get_path(
            'ext_test_resources/{0}'.format(obj_id)))
        obj2 = res['ext_test_resource']
        self.assertEqual(attr, obj2[extattr.EXTENDED_ATTRIBUTE])



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import timeutils

from tacker.common import exceptions
from tacker.common import utils
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import base
from tacker.objects import fields


LOG = logging.getLogger(__name__)


@db_api.context_manager.writer
def _destroy_instantiated_vnf_info(context, uuid):
    now = timeutils.utcnow()
    updated_values = {'deleted': True,
                      'deleted_at': now
                      }
    api.model_query(context, models.VnfInstantiatedInfo). \
        filter_by(vnf_instance_id=uuid). \
        update(updated_values, synchronize_session=False)


@db_api.context_manager.writer
def _instantiate_vnf_info_update(context, vnf_instance_id, values):
    vnf_info = api.model_query(context, models.VnfInstantiatedInfo). \
        filter_by(vnf_instance_id=vnf_instance_id).first()

    needs_create = False
    if vnf_info and vnf_info['deleted']:
        raise exceptions.VnfInstantiatedInfoNotFound(
            vnf_instance_id=vnf_instance_id)
    elif not vnf_info:
        values['vnf_instance_id'] = vnf_instance_id
        vnf_info = models.VnfInstantiatedInfo(**values)
        needs_create = True

    if needs_create:
        vnf_info.save(session=context.session)
    else:
        vnf_info.update(values)
        vnf_info.save(session=context.session)

    return vnf_info


@base.TackerObjectRegistry.register
class InstantiatedVnfInfo(base.TackerObject, base.TackerObjectDictCompat,
                     base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'flavour_id': fields.StringField(nullable=False),
        'vnf_instance_id': fields.UUIDField(nullable=False),
        'scale_status': fields.ListOfObjectsField(
            'ScaleInfo', nullable=True, default=[]),
        'ext_cp_info': fields.ListOfObjectsField(
            'VnfExtCpInfo', nullable=False),
        'ext_virtual_link_info': fields.ListOfObjectsField(
            'ExtVirtualLinkInfo', nullable=True, default=[]),
        'ext_managed_virtual_link_info': fields.ListOfObjectsField(
            'ExtManagedVirtualLinkInfo', nullable=True, default=[]),
        'vnfc_resource_info': fields.ListOfObjectsField(
            'VnfcResourceInfo', nullable=True, default=[]),
        'vnf_virtual_link_resource_info': fields.ListOfObjectsField(
            'VnfVirtualLinkResourceInfo', nullable=True, default=[]),
        'virtual_storage_resource_info': fields.ListOfObjectsField(
            'VirtualStorageResourceInfo', nullable=True, default=[]),
        'vnfc_info': fields.ListOfObjectsField(
            'VnfcInfo', nullable=True, default=[]),
        'vnf_state': fields.VnfOperationalStateTypeField(nullable=False,
            default=fields.VnfOperationalStateType.STOPPED),
        'instance_id': fields.StringField(nullable=True, default=None),
        'instantiation_level_id': fields.StringField(nullable=True,
            default=None),
        'additional_params': fields.DictOfNullableField(nullable=True,
                                                       default={})
    }

    ALL_ATTRIBUTES = {
        'instantiatedInfo': {
            'flavourId': ('id', 'string', 'VnfInstantiatedInfo'),
            'vnfInstanceId':
                ('vnf_instance_id', 'uuid', 'VnfInstantiatedInfo'),
            'vnfState': ('vnf_state', 'string', 'VnfInstantiatedInfo'),
            'instanceId': ('instance_id', 'string', 'VnfInstantiatedInfo'),
            'instantiationLevelId':
                ('instantiation_level_id', 'string', 'VnfInstantiatedInfo'),
            'extCpInfo/*': ('ext_cp_info', 'key_value_pair',
                            {"key_column": "key", "value_column": "value",
                             "model": "VnfInstantiatedInfo"}),
            'extVirtualLinkInfo/*': ('ext_virtual_link_info', 'key_value_pair',
                                {"key_column": "key", "value_column": "value",
                                "model": "VnfInstantiatedInfo"}),
            'extManagedVirtualLinkInfo/*': (
                'ext_managed_virtual_link_info', 'key_value_pair',
                {"key_column": "key", "value_column": "value",
                "model": "VnfInstantiatedInfo"}),
            'vnfcResourceInfo/*': (
                'vnfc_resource_info', 'key_value_pair',
                {"key_column": "key", "value_column": "value",
                "model": "VnfInstantiatedInfo"}),
            'vnfVirtualLinkResourceInfo/*': (
                'vnf_virtual_link_resource_info', 'key_value_pair',
                {"key_column": "key", "value_column": "value",
                "model": "VnfInstantiatedInfo"}),
            'virtualStorageResourceInfo/*': (
                'virtual_storage_resource_info', 'key_value_pair',
                {"key_column": "key", "value_column": "value",
                "model": "VnfInstantiatedInfo"}),
            'additionalParams/*': (
                'additional_params', 'key_value_pair',
                {"key_column": "key", "value_column": "value",
                "model": "VnfInstantiatedInfo"}),
            'vnfcInfo/*': (
                'vnfc_info', 'key_value_pair',
                {"key_column": "key", "value_column": "value",
                "model": "VnfInstantiatedInfo"}),
        }
    }

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())

    @staticmethod
    def _from_db_object(context, inst_vnf_info, db_inst_vnf_info):

        special_fields = ['scale_status',
                          'ext_cp_info', 'ext_virtual_link_info',
                          'ext_managed_virtual_link_info',
                          'vnfc_resource_info',
                          'vnf_virtual_link_resource_info',
                          'virtual_storage_resource_info',
                          'vnfc_info']
        for key in inst_vnf_info.fields:
            if key in special_fields:
                continue

            setattr(inst_vnf_info, key, db_inst_vnf_info.get(key))

        scale_status = db_inst_vnf_info['scale_status']
        scale_status_list = [ScaleInfo.obj_from_primitive(scale, context)
                    for scale in scale_status]
        inst_vnf_info.scale_status = scale_status_list

        ext_cp_info = db_inst_vnf_info['ext_cp_info']
        ext_cp_info_list = [VnfExtCpInfo.obj_from_primitive(ext_cp, context)
                    for ext_cp in ext_cp_info]
        inst_vnf_info.ext_cp_info = ext_cp_info_list

        vnfc_resource_info = db_inst_vnf_info['vnfc_resource_info']
        vnfc_resource_info_list = [VnfcResourceInfo.obj_from_primitive(
            vnfc_resource, context) for vnfc_resource in vnfc_resource_info]
        inst_vnf_info.vnfc_resource_info = vnfc_resource_info_list

        storage_res_info = db_inst_vnf_info['virtual_storage_resource_info']
        storage_res_info_list = [VirtualStorageResourceInfo.
            obj_from_primitive(storage_resource, context)
            for storage_resource in storage_res_info]
        inst_vnf_info.virtual_storage_resource_info = storage_res_info_list

        ext_virtual_link_info = db_inst_vnf_info['ext_virtual_link_info']
        ext_vl_info_list = [ExtVirtualLinkInfo.obj_from_primitive(
            ext_vl_info, context) for ext_vl_info in ext_virtual_link_info]
        inst_vnf_info.ext_virtual_link_info = ext_vl_info_list

        ext_mng_vl_info = db_inst_vnf_info['ext_managed_virtual_link_info']
        ext_managed_vl_info_list = [ExtManagedVirtualLinkInfo.
            obj_from_primitive(ext_managed_vl_info, context) for
            ext_managed_vl_info in ext_mng_vl_info]
        inst_vnf_info.ext_managed_virtual_link_info = ext_managed_vl_info_list

        vnf_vl_resource_info = db_inst_vnf_info[
            'vnf_virtual_link_resource_info']
        vnf_vl_info_list = [VnfVirtualLinkResourceInfo.
            obj_from_primitive(vnf_vl_info, context) for vnf_vl_info in
            vnf_vl_resource_info]
        inst_vnf_info.vnf_virtual_link_resource_info = vnf_vl_info_list

        vnfc_info = db_inst_vnf_info[
            'vnfc_info']
        vnfc_info_list = [VnfcInfo.
            obj_from_primitive(vnfc, context) for vnfc in
            vnfc_info]
        inst_vnf_info.vnfc_info = vnfc_info_list

        inst_vnf_info._context = context
        inst_vnf_info.obj_reset_changes()
        return inst_vnf_info

    @base.remotable
    def save(self):
        updates = {}
        changes = self.obj_what_changed()

        for field in self.fields:
            if (self.obj_attr_is_set(field) and
                    isinstance(self.fields[field], fields.ListOfObjectsField)):
                field_list = getattr(self, field)
                updates[field] = [obj.obj_to_primitive() for obj in field_list]
            elif field in changes:
                updates[field] = self[field]

        vnf_info = _instantiate_vnf_info_update(self._context,
                                                self.vnf_instance_id,
                                                updates)
        self._from_db_object(self._context, self, vnf_info)

        self.obj_reset_changes()

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            instantiate_vnf_info = super(
                InstantiatedVnfInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'scale_status' in primitive.keys():
                obj_data = [ScaleInfo.obj_from_primitive(
                    scale, context) for scale in primitive.get(
                    'scale_status', [])]
                primitive.update({'scale_status': obj_data})

            if 'ext_cp_info' in primitive.keys():
                obj_data = [VnfExtCpInfo.obj_from_primitive(
                    vnf_ext_cp, context) for vnf_ext_cp in primitive.get(
                    'ext_cp_info', [])]
                primitive.update({'ext_cp_info': obj_data})

            if 'ext_virtual_link_info' in primitive.keys():
                obj_data = [ExtVirtualLinkInfo.obj_from_primitive(
                    ext_virtual_link,
                    context) for ext_virtual_link in primitive.get(
                    'ext_virtual_link_info', [])]
                primitive.update({'ext_virtual_link_info': obj_data})

            if 'ext_managed_virtual_link_info' in primitive.keys():
                obj_data = [ExtManagedVirtualLinkInfo.obj_from_primitive(
                    ext_managed_v_link,
                    context) for ext_managed_v_link in primitive.get(
                    'ext_managed_virtual_link_info', [])]
                primitive.update({'ext_managed_virtual_link_info': obj_data})

            if 'vnfc_resource_info' in primitive.keys():
                obj_data = [VnfcResourceInfo.obj_from_primitive(
                    vnf_resource_info,
                    context) for vnf_resource_info in primitive.get(
                    'vnfc_resource_info', [])]
                primitive.update({'vnfc_resource_info': obj_data})

            if 'vnf_virtual_link_resource_info' in primitive.keys():
                obj_data = [VnfVirtualLinkResourceInfo.obj_from_primitive(
                    vnf_v_link_resource,
                    context) for vnf_v_link_resource in primitive.get(
                    'vnf_virtual_link_resource_info', [])]
                primitive.update({'vnf_virtual_link_resource_info': obj_data})

            if 'virtual_storage_resource_info' in primitive.keys():
                obj_data = [VirtualStorageResourceInfo.obj_from_primitive(
                    virtual_storage_info,
                    context) for virtual_storage_info in primitive.get(
                    'virtual_storage_resource_info', [])]
                primitive.update({'virtual_storage_resource_info': obj_data})

            if 'vnfc_info' in primitive.keys():
                obj_data = [VnfcInfo.obj_from_primitive(
                    vnfc_info, context) for vnfc_info in primitive.get(
                    'vnfc_info', [])]
                primitive.update({'vnfc_info': obj_data})

            instantiate_vnf_info = \
                InstantiatedVnfInfo._from_dict(primitive)

        return instantiate_vnf_info

    @classmethod
    def obj_from_db_obj(cls, context, db_obj):
        return cls._from_db_object(context, cls(), db_obj)

    @classmethod
    def _from_dict(cls, data_dict):
        flavour_id = data_dict.get('flavour_id')
        scale_status = data_dict.get('scale_status', [])
        ext_cp_info = data_dict.get('ext_cp_info', [])
        ext_virtual_link_info = data_dict.get('ext_virtual_link_info', [])
        ext_managed_virtual_link_info = data_dict.get(
            'ext_managed_virtual_link_info', [])
        vnfc_resource_info = data_dict.get('vnfc_resource_info', [])
        vnf_virtual_link_resource_info = data_dict.get(
            'vnf_virtual_link_resource_info', [])
        virtual_storage_resource_info = data_dict.get(
            'virtual_storage_resource_info', [])
        vnf_state = data_dict.get('vnf_state')
        instantiation_level_id = data_dict.get('instantiation_level_id')
        additional_params = data_dict.get('additional_params', {})
        vnfc_info = data_dict.get('vnfc_info', [])

        obj = cls(flavour_id=flavour_id,
               scale_status=scale_status,
               ext_cp_info=ext_cp_info,
               ext_virtual_link_info=ext_virtual_link_info,
               ext_managed_virtual_link_info=ext_managed_virtual_link_info,
               vnfc_resource_info=vnfc_resource_info,
               vnf_virtual_link_resource_info=vnf_virtual_link_resource_info,
               virtual_storage_resource_info=virtual_storage_resource_info,
               vnfc_info=vnfc_info,
               vnf_state=vnf_state,
               instantiation_level_id=instantiation_level_id,
               additional_params=additional_params)
        return obj

    def to_dict(self):
        data = {'flavour_id': self.flavour_id,
            'vnf_state': self.vnf_state}

        if self.scale_status:
            scale_status_list = []
            for scale_status in self.scale_status:
                scale_status_list.append(scale_status.to_dict())

            data.update({'scale_status': scale_status_list})

        ext_cp_info_list = []
        for ext_cp_info in self.ext_cp_info:
            ext_cp_info_list.append(ext_cp_info.to_dict())

        data.update({'ext_cp_info': ext_cp_info_list})

        if self.ext_virtual_link_info:
            exp_virt_link_info_list = []
            for exp_virt_link_info in self.ext_virtual_link_info:
                exp_virt_link_info_list.append(exp_virt_link_info.to_dict())
            data.update({'ext_virtual_link_info': exp_virt_link_info_list})

        if self.ext_managed_virtual_link_info:
            ext_managed_virt_info_list = []
            for exp_managed_virt_link_info in \
                    self.ext_managed_virtual_link_info:
                info = exp_managed_virt_link_info.to_dict()
                ext_managed_virt_info_list.append(info)
            data.update({'ext_managed_virtual_link_info':
                    ext_managed_virt_info_list})

        if self.vnfc_resource_info:
            vnfc_resource_info_list = []
            for vnfc_resource_info in self.vnfc_resource_info:
                vnfc_resource_info_list.append(vnfc_resource_info.to_dict())

            data.update({'vnfc_resource_info': vnfc_resource_info_list})

        if self.vnf_virtual_link_resource_info:
            virt_link_info = []
            for vnf_virtual_link_resource_info in \
                    self.vnf_virtual_link_resource_info:
                info = vnf_virtual_link_resource_info.to_dict()
                virt_link_info.append(info)

            data.update({'vnf_virtual_link_resource_info': virt_link_info})

        if self.virtual_storage_resource_info:
            virtual_storage_resource_info_list = []
            for virtual_storage_resource_info in \
                    self.virtual_storage_resource_info:
                info = virtual_storage_resource_info.to_dict()
                virtual_storage_resource_info_list.append(info)

            data.update({'virtual_storage_resource_info':
                    virtual_storage_resource_info_list})

        if self.vnfc_info:
            vnfc_info = []
            for vnfc in self.vnfc_info:
                info = vnfc.to_dict()
                vnfc_info.append(info)

            data.update({'vnfc_info': vnfc_info})

        data.update({'additional_params':
                self.additional_params})

        return data

    def reinitialize(self):
        # Reinitialize vnf to non instantiated state.
        self.scale_status = []
        self.ext_cp_info = []
        self.ext_virtual_link_info = []
        self.ext_managed_virtual_link_info = []
        self.vnfc_resource_info = []
        self.vnf_virtual_link_resource_info = []
        self.virtual_storage_resource_info = []
        self.instance_id = None
        self.vnf_state = fields.VnfOperationalStateType.STOPPED
        self.vnfc_info = []

    @base.remotable
    def destroy(self, context):
        if not self.obj_attr_is_set('vnf_instance_id'):
            raise exceptions.ObjectActionError(action='destroy',
                                               reason='no uuid')

        _destroy_instantiated_vnf_info(context, self.vnf_instance_id)


@base.TackerObjectRegistry.register
class ScaleInfo(base.TackerObject, base.TackerObjectDictCompat,
                base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'aspect_id': fields.StringField(nullable=False),
        'scale_level': fields.IntegerField(nullable=False),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_scale_status = super(
                ScaleInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            obj_scale_status = ScaleInfo._from_dict(primitive)

        return obj_scale_status

    @classmethod
    def _from_dict(cls, data_dict):
        aspect_id = data_dict.get('aspect_id')
        scale_level = data_dict.get('scale_level')

        obj = cls(aspect_id=aspect_id,
                  scale_level=scale_level)
        return obj

    def to_dict(self):

        return {'aspect_id': self.aspect_id,
                'scale_level': self.scale_level}


@base.TackerObjectRegistry.register
class VnfExtCpInfo(base.TackerObject, base.TackerObjectDictCompat,
                   base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'cpd_id': fields.StringField(nullable=False),
        'cp_protocol_info': fields.ListOfObjectsField(
            'CpProtocolInfo', nullable=False, default=[]),
        'ext_link_port_id': fields.StringField(nullable=True, default=None),
        'associated_vnfc_cp_id': fields.StringField(nullable=True)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_ext_cp_info = super(
                VnfExtCpInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'cp_protocol_info' in primitive.keys():
                obj_data = [CpProtocolInfo.obj_from_primitive(
                    ext_cp, context) for ext_cp in primitive.get(
                    'cp_protocol_info', [])]
                primitive.update({'cp_protocol_info': obj_data})

            obj_ext_cp_info = VnfExtCpInfo._from_dict(primitive)

        return obj_ext_cp_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        cpd_id = data_dict.get('cpd_id')
        cp_protocol_info = data_dict.get('cp_protocol_info', [])
        ext_link_port_id = data_dict.get('ext_link_port_id')
        associated_vnfc_cp_id = data_dict.get('associated_vnfc_cp_id')

        obj = cls(id=id, cpd_id=cpd_id,
                cp_protocol_info=cp_protocol_info,
                ext_link_port_id=ext_link_port_id,
                associated_vnfc_cp_id=associated_vnfc_cp_id)
        return obj

    def to_dict(self):
        data = {'id': self.id,
            'cpd_id': self.cpd_id,
            'ext_link_port_id': self.ext_link_port_id,
            'associated_vnfc_cp_id': self.associated_vnfc_cp_id}

        cp_protocol_info_list = []
        for cp_protocol_info in self.cp_protocol_info:
            cp_protocol_info_list.append(cp_protocol_info.to_dict())

        data.update({'cp_protocol_info': cp_protocol_info_list})

        return data


@base.TackerObjectRegistry.register
class CpProtocolInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'layer_protocol': fields.StringField(nullable=False),
        'ip_over_ethernet': fields.ObjectField(
            'IpOverEthernetAddressInfo', nullable=True, default=None),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_cp_protocol = super(CpProtocolInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'ip_over_ethernet' in primitive.keys():
                obj_data = IpOverEthernetAddressInfo.obj_from_primitive(
                    primitive.get('ip_over_ethernet', {}), context)
                primitive.update({'ip_over_ethernet': obj_data})

            obj_cp_protocol = CpProtocolInfo._from_dict(primitive)

        return obj_cp_protocol

    @classmethod
    def _from_dict(cls, data_dict):
        layer_protocol = data_dict.get('layer_protocol')
        ip_over_ethernet = data_dict.get('ip_over_ethernet')

        obj = cls(layer_protocol=layer_protocol,
                  ip_over_ethernet=ip_over_ethernet)
        return obj

    def to_dict(self):
        data = {'layer_protocol': self.layer_protocol}

        if self.ip_over_ethernet:
            data.update({'ip_over_ethernet': self.ip_over_ethernet.to_dict()})

        return data


@base.TackerObjectRegistry.register
class IpOverEthernetAddressInfo(base.TackerObject,
                                base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'mac_address': fields.StringField(nullable=True, default=None),
        'ip_addresses': fields.ListOfObjectsField('IpAddress', nullable=True,
                                                  default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            ip_over_ethernet = super(
                IpOverEthernetAddressInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'ip_addresses' in primitive.keys():
                obj_data = [IpAddress._from_dict(
                    ip_address) for ip_address in primitive.get(
                        'ip_addresses', [])]
                primitive.update({'ip_addresses': obj_data})

            ip_over_ethernet = IpOverEthernetAddressInfo._from_dict(primitive)

        return ip_over_ethernet

    @classmethod
    def _from_dict(cls, data_dict):
        mac_address = data_dict.get('mac_address')
        ip_addresses = data_dict.get('ip_addresses', [])
        obj = cls(mac_address=mac_address, ip_addresses=ip_addresses)
        return obj

    def to_dict(self):
        data = {'mac_address': self.mac_address}

        if self.ip_addresses:
            ip_addresses_list = []
            for ip_addresses in self.ip_addresses:
                ip_addresses_list.append(ip_addresses.to_dict())

            data.update({'ip_addresses': ip_addresses_list})

        return data


@base.TackerObjectRegistry.register
class IpAddress(base.TackerObject, base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'type': fields.IpAddressTypeField(nullable=False),
        'subnet_id': fields.StringField(nullable=True, default=None),
        'is_dynamic': fields.BooleanField(nullable=True, default=False),
        'addresses': fields.ListOfStringsField(nullable=True, default=[]),
    }

    @classmethod
    def _from_dict(cls, data_dict):
        type = data_dict.get('type', fields.IpAddressType.IPV4)
        subnet_id = data_dict.get('subnet_id')
        is_dynamic = data_dict.get('is_dynamic', False)
        addresses = data_dict.get('addresses', [])

        obj = cls(type=type, subnet_id=subnet_id, is_dynamic=is_dynamic,
                  addresses=addresses)

        return obj

    def to_dict(self):
        return {'type': self.type,
            'subnet_id': self.subnet_id,
            'is_dynamic': self.is_dynamic,
            'addresses': self.addresses}


@base.TackerObjectRegistry.register
class ExtVirtualLinkInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'resource_handle': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'ext_link_ports': fields.ListOfObjectsField(
            'ExtLinkPortInfo', nullable=True, default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_ext_virt_link = super(
                ExtVirtualLinkInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'resource_handle' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('resource_handle'))
                primitive.update({'resource_handle': obj_data})

            if 'ext_link_ports' in primitive.keys():
                obj_data = [ExtLinkPortInfo.obj_from_primitive(
                    ext_link_port_info, context)
                    for ext_link_port_info in primitive.get(
                        'ext_link_ports', [])]
                primitive.update({'ext_link_ports': obj_data})

            obj_ext_virt_link = ExtVirtualLinkInfo._from_dict(primitive)

        return obj_ext_virt_link

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id', '')
        resource_handle = data_dict.get('resource_handle')
        ext_link_ports = data_dict.get('ext_link_ports', [])

        obj = cls(id=id, resource_handle=resource_handle,
                  ext_link_ports=ext_link_ports)
        return obj

    def to_dict(self):
        data = {'id': self.id,
            'resource_handle': self.resource_handle.to_dict()}

        if self.ext_link_ports:
            ext_link_ports = []
            for ext_link_port in self.ext_link_ports:
                ext_link_ports.append(ext_link_port.to_dict())

            data.update({'ext_link_ports': ext_link_ports})

        return data


@base.TackerObjectRegistry.register
class ExtLinkPortInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'resource_handle': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'cp_instance_id': fields.StringField(nullable=True, default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_link_port_info = super(
                ExtLinkPortInfo, cls).obj_from_primitive(primitive, context)
        else:
            if 'resource_handle' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('resource_handle'))
                primitive.update({'resource_handle': obj_data})

            obj_link_port_info = ExtLinkPortInfo._from_dict(primitive)

        return obj_link_port_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        resource_handle = data_dict.get('resource_handle')
        cp_instance_id = data_dict.get('cp_instance_id')
        obj = cls(id=id, resource_handle=resource_handle,
                  cp_instance_id=cp_instance_id)
        return obj

    def to_dict(self):
        return {'id': self.id,
            'resource_handle': self.resource_handle.to_dict(),
            'cp_instance_id': self.cp_instance_id}


@base.TackerObjectRegistry.register
class ExtManagedVirtualLinkInfo(base.TackerObject,
                                base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnf_virtual_link_desc_id': fields.StringField(nullable=False),
        'network_resource': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'vnf_link_ports': fields.ListOfObjectsField(
            'VnfLinkPortInfo', nullable=True, default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_ext_managed_virt_link = super(
                ExtManagedVirtualLinkInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'network_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('network_resource'))
                primitive.update({'network_resource': obj_data})

            if 'vnf_link_ports' in primitive.keys():
                obj_data = [VnfLinkPortInfo.obj_from_primitive(
                    vnf_link_port, context)
                    for vnf_link_port in primitive.get(
                        'vnf_link_ports', [])]
                primitive.update({'vnf_link_ports': obj_data})

            obj_ext_managed_virt_link = ExtManagedVirtualLinkInfo._from_dict(
                primitive)

        return obj_ext_managed_virt_link

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vnf_virtual_link_desc_id = data_dict.get('vnf_virtual_link_desc_id')
        network_resource = data_dict.get('network_resource')
        vnf_link_ports = data_dict.get('vnf_link_ports', [])

        obj = cls(id=id, vnf_virtual_link_desc_id=vnf_virtual_link_desc_id,
                  network_resource=network_resource,
                  vnf_link_ports=vnf_link_ports)
        return obj

    def to_dict(self):
        data = {'id': self.id,
            'vnf_virtual_link_desc_id': self.vnf_virtual_link_desc_id,
            'network_resource': self.network_resource.to_dict()}

        if self.vnf_link_ports:
            vnf_link_ports = []
            for vnf_link_port in self.vnf_link_ports:
                vnf_link_ports.append(vnf_link_port.to_dict())

            data.update({'vnf_link_ports': vnf_link_ports})

        return data


@base.TackerObjectRegistry.register
class VnfLinkPortInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'resource_handle': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'cp_instance_id': fields.StringField(nullable=True, default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            vnf_port_link_info = super(
                VnfLinkPortInfo, cls).obj_from_primitive(primitive, context)
        else:
            if 'resource_handle' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('resource_handle'))
                primitive.update({'resource_handle': obj_data})

            vnf_port_link_info = VnfLinkPortInfo._from_dict(primitive)

        return vnf_port_link_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        resource_handle = data_dict.get('resource_handle')
        cp_instance_id = data_dict.get('cp_instance_id')

        obj = cls(id=id, resource_handle=resource_handle,
                  cp_instance_id=cp_instance_id)
        return obj

    def to_dict(self):
        return {'id': self.id,
            'resource_handle': self.resource_handle.to_dict(),
            'cp_instance_id': self.cp_instance_id}


@base.TackerObjectRegistry.register
class VnfcResourceInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vdu_id': fields.StringField(nullable=False),
        'compute_resource': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'storage_resource_ids': fields.ListOfStringsField(nullable=True,
                                                          default=[]),
        'vnfc_cp_info': fields.ListOfObjectsField(
            'VnfcCpInfo', nullable=True, default=[]),
        'metadata': fields.DictOfStringsField(nullable=True, default={})

    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            vnfc_resource_info = super(
                VnfcResourceInfo, cls).obj_from_primitive(primitive, context)
        else:
            if 'compute_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('compute_resource'))
                primitive.update({'compute_resource': obj_data})

            if 'vnfc_cp_info' in primitive.keys():
                obj_data = [VnfcCpInfo.obj_from_primitive(
                    vnfc_cp_info, context)
                    for vnfc_cp_info in primitive.get(
                        'vnfc_cp_info', [])]
                primitive.update({'vnfc_cp_info': obj_data})

            vnfc_resource_info = VnfcResourceInfo._from_dict(primitive)

        return vnfc_resource_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vdu_id = data_dict.get('vdu_id')
        compute_resource = data_dict.get('compute_resource')
        storage_resource_ids = data_dict.get('storage_resource_ids', [])
        vnfc_cp_info = data_dict.get('vnfc_cp_info', [])
        metadata = data_dict.get('metadata', {})

        obj = cls(id=id, vdu_id=vdu_id,
                  compute_resource=compute_resource,
                  storage_resource_ids=storage_resource_ids,
                  vnfc_cp_info=vnfc_cp_info, metadata=metadata)

        return obj

    def to_dict(self):
        data = {'id': self.id,
            'vdu_id': self.vdu_id,
            'compute_resource': self.compute_resource.to_dict(),
            'storage_resource_ids': self.storage_resource_ids}

        if self.vnfc_cp_info:
            vnfc_cp_info_list = []
            for vnfc_cp_info in self.vnfc_cp_info:
                vnfc_cp_info_list.append(vnfc_cp_info.to_dict())

            data.update({'vnfc_cp_info': vnfc_cp_info_list})

        return data


@base.TackerObjectRegistry.register
class VnfcCpInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'cpd_id': fields.StringField(nullable=False),
        'vnf_ext_cp_id': fields.StringField(nullable=True, default=None),
        'cp_protocol_info': fields.ListOfObjectsField(
            'CpProtocolInfo', nullable=True, default=[]),
        'vnf_link_port_id': fields.StringField(nullable=True, default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_vnfc_cp_info = super(VnfcCpInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'cp_protocol_info' in primitive.keys():
                obj_data = [CpProtocolInfo.obj_from_primitive(
                    ext_cp, context) for ext_cp in primitive.get(
                    'cp_protocol_info', [])]
                primitive.update({'cp_protocol_info': obj_data})

            obj_vnfc_cp_info = VnfcCpInfo._from_dict(primitive)

        return obj_vnfc_cp_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        cpd_id = data_dict.get('cpd_id')
        vnf_ext_cp_id = data_dict.get('vnf_ext_cp_id')
        cp_protocol_info = data_dict.get('cp_protocol_info', [])
        vnf_link_port_id = data_dict.get('vnf_link_port_id')

        obj = cls(id=id, cpd_id=cpd_id,
                  vnf_ext_cp_id=vnf_ext_cp_id,
                  cp_protocol_info=cp_protocol_info,
                  vnf_link_port_id=vnf_link_port_id)

        return obj

    def to_dict(self):
        data = {'id': self.id,
             'cpd_id': self.cpd_id,
             'vnf_ext_cp_id': self.vnf_ext_cp_id,
             'vnf_link_port_id': self.vnf_link_port_id}

        if self.cp_protocol_info:
            cp_protocol_info_list = []
            for cp_protocol_info in self.cp_protocol_info:
                cp_protocol_info_list.append(cp_protocol_info.to_dict())

            data.update({'cp_protocol_info': cp_protocol_info_list})

        return data


@base.TackerObjectRegistry.register
class VnfVirtualLinkResourceInfo(base.TackerObject,
                                base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnf_virtual_link_desc_id': fields.StringField(nullable=False),
        'network_resource': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'vnf_link_ports': fields.ListOfObjectsField(
            'VnfLinkPortInfo', nullable=True, default=[])
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_vnf_virtual_link = super(
                VnfVirtualLinkResourceInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'network_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('network_resource'))
                primitive.update({'network_resource': obj_data})

            if 'vnf_link_ports' in primitive.keys():
                obj_data = [VnfLinkPortInfo.obj_from_primitive(
                    vnf_link_port, context)
                    for vnf_link_port in primitive.get(
                        'vnf_link_ports', [])]
                primitive.update({'vnf_link_ports': obj_data})

            obj_vnf_virtual_link = VnfVirtualLinkResourceInfo._from_dict(
                primitive)

        return obj_vnf_virtual_link

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vnf_virtual_link_desc_id = data_dict.get(
            'vnf_virtual_link_desc_id')
        network_resource = data_dict.get('network_resource')
        vnf_link_ports = data_dict.get('vnf_link_ports', [])

        obj = cls(id=id, vnf_virtual_link_desc_id=vnf_virtual_link_desc_id,
                  network_resource=network_resource,
                  vnf_link_ports=vnf_link_ports)

        return obj

    def to_dict(self):
        data = {'id': self.id,
            'vnf_virtual_link_desc_id': self.vnf_virtual_link_desc_id,
            'network_resource': self.network_resource.to_dict()}

        if self.vnf_link_ports:
            vnf_link_ports = []
            for vnf_link_port in self.vnf_link_ports:
                vnf_link_ports.append(vnf_link_port.to_dict())

            data['vnf_link_ports'] = vnf_link_ports

        return data


@base.TackerObjectRegistry.register
class VirtualStorageResourceInfo(base.TackerObject,
                                base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'virtual_storage_desc_id': fields.StringField(nullable=False),
        'storage_resource': fields.ObjectField(
            'ResourceHandle', nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_virtual_storage = super(
                VirtualStorageResourceInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'storage_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('storage_resource'))
                primitive.update({'storage_resource': obj_data})

            obj_virtual_storage = VirtualStorageResourceInfo._from_dict(
                primitive)

        return obj_virtual_storage

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        virtual_storage_desc_id = data_dict.get('virtual_storage_desc_id')
        storage_resource = data_dict.get('storage_resource')

        obj = cls(id=id, virtual_storage_desc_id=virtual_storage_desc_id,
                  storage_resource=storage_resource)

        return obj

    def to_dict(self):
        return {'id': self.id,
            'virtual_storage_desc_id': self.virtual_storage_desc_id,
            'storage_resource': self.storage_resource.to_dict()}


@base.TackerObjectRegistry.register
class VnfcInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vdu_id': fields.StringField(nullable=False),
        'vnfc_state': fields.StringField(nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_vnfc_info = super(
                VnfcInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            obj_vnfc_info = VnfcInfo._from_dict(
                primitive)

        return obj_vnfc_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vdu_id = data_dict.get('vdu_id')
        vnfc_state = data_dict.get('vnfc_state')

        obj = cls(id=id, vdu_id=vdu_id,
                  vnfc_state=vnfc_state)

        return obj

    def to_dict(self):
        return {'id': self.id,
            'vdu_id': self.vdu_id,
            'vnfc_state': self.vnfc_state}


@base.TackerObjectRegistry.register
class ResourceHandle(base.TackerObject,
                     base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vim_connection_id': fields.StringField(nullable=True,
                                                default=None),
        'resource_id': fields.StringField(nullable=False, default=""),
        'vim_level_resource_type': fields.StringField(nullable=True,
                                                      default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            resource_handle = super(
                ResourceHandle, cls).obj_from_primitive(
                primitive, context)
        else:
            resource_handle = ResourceHandle._from_dict(primitive)

        return resource_handle

    @classmethod
    def _from_dict(cls, data_dict):
        vim_connection_id = data_dict.get('vim_connection_id')
        resource_id = data_dict.get('resource_id', "")
        vim_level_resource_type = data_dict.get('vim_level_resource_type')

        obj = cls(vim_connection_id=vim_connection_id,
                  resource_id=resource_id,
                  vim_level_resource_type=vim_level_resource_type)

        return obj

    def to_dict(self):
        return {'vim_connection_id': self.vim_connection_id,
                'resource_id': self.resource_id,
                'vim_level_resource_type': self.vim_level_resource_type}



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os

from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.common import utils as common_utils
import tacker.conf
from tacker.objects import fields
from tacker.objects import vnf_package
from tacker.objects import vnf_package_vnfd

from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored.common import vim_utils
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored.common import vnfd_utils
from tacker.sol_refactored.nfvo import glance_utils
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields as v2_fields


CONF = tacker.conf.CONF

LOG = logging.getLogger(__name__)

# NOTE:
# It is NFVO implementation used when an external NFVO is not used.
# It implements only functions necessary for vnflcm v2.
# It uses original tacker vnfpkgm v1 for vnf package managemnet
# and adds grant functions.


class LocalNfvo(object):

    def __init__(self):
        self.inst_vim_info = {}
        self.inst_vnfd_id = {}

    def onboarded_show(self, context, id):
        pkg_vnfd = vnf_package_vnfd.VnfPackageVnfd().get_by_vnfdId(
            context, id)
        if not pkg_vnfd:
            raise sol_ex.VnfdIdNotFound(vnfd_id=id)

        vnf_pkg = vnf_package.VnfPackage().get_by_id(
            context, pkg_vnfd.package_uuid)
        if not vnf_pkg:
            # never happen. just for code consistency.
            raise sol_ex.VnfdIdNotFound(vnfd_id=id)

        if (vnf_pkg.onboarding_state !=
                fields.PackageOnboardingStateType.ONBOARDED):
            # NOTE: API itself should be return 409 but it is used by
            # only vnf_instance create and will be converted to 422 in it.
            raise sol_ex.VnfdIdNotFound(vnfd_id=id)

        # NOTE:
        # This is used for vnf_instance create at the moment.
        # It is set only attributes necessary for vnf_instance create.
        res = objects.VnfPkgInfoV2(
            id=pkg_vnfd.package_uuid,
            vnfdId=pkg_vnfd.vnfd_id,
            vnfProvider=pkg_vnfd.vnf_provider,
            vnfProductName=pkg_vnfd.vnf_product_name,
            vnfSoftwareVersion=pkg_vnfd.vnf_software_version,
            vnfdVersion=pkg_vnfd.vnfd_version,
            operationalState=vnf_pkg.operational_state
        )

        return res

    def get_csar_dir(self, context, vnfd_id):
        pkg_vnfd = vnf_package_vnfd.VnfPackageVnfd().get_by_vnfdId(
            context, vnfd_id)
        if not pkg_vnfd:
            raise sol_ex.VnfdIdNotFound(vnfd_id=vnfd_id)

        csar_dir = os.path.join(CONF.vnf_package.vnf_package_csar_path,
                                pkg_vnfd.package_uuid)
        if not os.path.isdir(csar_dir):
            raise sol_ex.VnfdIdNotFound(vnfd_id=vnfd_id)

        return csar_dir

    def get_vnfd(self, context, vnfd_id):
        vnfd = vnfd_utils.Vnfd(vnfd_id)
        vnfd.init_from_csar_dir(self.get_csar_dir(context, vnfd_id))
        return vnfd

    def _glance_create_image(self, vim_info, vnfd, sw_data, inst_id):
        min_disk = 0
        if 'min_disk' in sw_data:
            min_disk = common_utils.MemoryUnit.convert_unit_size_to_num(
                sw_data['min_disk'], 'GB')

        min_ram = 0
        if 'min_ram' in sw_data:
            min_ram = common_utils.MemoryUnit.convert_unit_size_to_num(
                sw_data['min_ram'], 'MB')

        filename = os.path.join(vnfd.csar_dir,
                                sw_data['file'].split('../')[-1])

        # NOTE: use tag to find to delete images when terminate vnf instance.
        create_args = {
            'min_disk': min_disk,
            'min_ram': min_ram,
            'disk_format': sw_data.get('disk_format'),
            'container_format': sw_data.get('container_format'),
            'filename': filename,
            'visibility': 'private',
            'tags': [inst_id]
        }

        glance_client = glance_utils.GlanceClient(vim_info)
        image = glance_client.create_image(sw_data['name'], **create_args)

        LOG.debug("image created name: %s id: %s", sw_data['name'], image.id)

        return image.id

    def _get_vim_info(self, context, grant_req):
        lcmocc = lcmocc_utils.get_lcmocc(context, grant_req.vnfLcmOpOccId)
        inst_req = lcmocc.operationParams
        if inst_req.obj_attr_is_set('vimConnectionInfo'):
            return inst_utils.select_vim_info(inst_req.vimConnectionInfo)

        # NOTE: exception is not raised.
        return vim_utils.get_default_vim(context)

    def instantiate_grant(self, context, grant_req, grant_res):
        # handle ZoneInfo
        zone_list = config.CONF.v2_nfvo.test_grant_zone_list
        zone_id = None
        if len(zone_list) > 0:
            zone_infos = []
            for zone in zone_list:
                zone_info = objects.ZoneInfoV1(
                    id=uuidutils.generate_uuid(),
                    zoneId=zone
                )
                zone_infos.append(zone_info)
            grant_res.zones = zone_infos
            zone_id = zone_infos[0].id

        # handle addResources.
        # only copy req to res. i.e. grant all.
        attr = 'addResources'
        if grant_req.obj_attr_is_set(attr):
            add_res = []
            for res_def in grant_req[attr]:
                g_info = objects.GrantInfoV1(
                    resourceDefinitionId=res_def.id
                )
                if zone_id is not None and res_def.type == 'COMPUTE':
                    g_info.zoneId = zone_id
                add_res.append(g_info)
            grant_res[attr] = add_res

        # handle vimAssets
        # if there is an artifact, create glance image.
        vnfd = self.get_vnfd(context, grant_req.vnfdId)
        sw_image_data = vnfd.get_sw_image_data(grant_req.flavourId)
        vim_sw_images = []
        for res_id, sw_data in sw_image_data.items():
            if 'file' in sw_data:
                # if artifact is specified, create glance image.
                # it may fail. catch exception and raise 403(not granted)
                # if error occur.

                # get vim_info to access glance
                vim_info = self._get_vim_info(context, grant_req)
                if vim_info is None:
                    msg = "No vimConnectionInfo to create glance image"
                    raise sol_ex.LocalNfvoGrantFailed(sol_detail=msg)

                try:
                    image = self._glance_create_image(vim_info, vnfd, sw_data,
                            grant_req.vnfInstanceId)
                except Exception:
                    msg = "glance image create failed"
                    LOG.exception(msg)
                    raise sol_ex.LocalNfvoGrantFailed(sol_detail=msg)
            else:
                # there is no artifact. suppose image already created.
                image = sw_data['name']
            vim_sw_image = objects.VimSoftwareImageV1(
                vnfdSoftwareImageId=res_id,
                vimSoftwareImageId=image)
            vim_sw_images.append(vim_sw_image)
        if vim_sw_images:
            grant_res.vimAssets = objects.GrantV1_VimAssets(
                softwareImages=vim_sw_images
            )

    def grant(self, context, grant_req):
        grant_res = objects.GrantV1(
            id=uuidutils.generate_uuid(),
            vnfInstanceId=grant_req.vnfInstanceId,
            vnfLcmOpOccId=grant_req.vnfLcmOpOccId
        )

        # NOTE: considered instantiate only at the moment.
        # terminate is granted with no grant_res constructed.
        if grant_req.operation == v2_fields.LcmOperationType.INSTANTIATE:
            self.instantiate_grant(context, grant_req, grant_res)

        endpoint = config.CONF.v2_vnfm.endpoint
        grant_res._links = objects.GrantV1_Links(
            vnfLcmOpOcc=objects.Link(
                href=lcmocc_utils.lcmocc_href(grant_req.vnfLcmOpOccId,
                                              endpoint)),
            vnfInstance=objects.Link(
                href=inst_utils.inst_href(grant_req.vnfInstanceId,
                                          endpoint))
        )
        grant_res._links.self = objects.Link(
            href="{}/grant/v1/grants/{}".format(endpoint, grant_res.id))

        return grant_res

    def _update_vnf_pkg_usage_state(self, context, vnfd_id, state):
        pkg_vnfd = vnf_package_vnfd.VnfPackageVnfd().get_by_vnfdId(
            context, vnfd_id)
        if not pkg_vnfd:
            # should not happen. just for code consistency.
            LOG.error("VnfPackage of vnfdID %s not found.", vnfd_id)
            return

        vnf_pkg = vnf_package.VnfPackage().get_by_id(
            context, pkg_vnfd.package_uuid)
        if not vnf_pkg:
            # should not happen. just for code consistency.
            LOG.error("VnfPackage %s not found.", pkg_vnfd.package_uuid)
            return

        # Multiple vnf instances can be created with same vnfd_id,
        # so the state must be changed to `NOT_IN_USE` only when
        # there is no vnf instance.
        if state == fields.PackageUsageStateType.NOT_IN_USE:
            insts = objects.VnfInstanceV2.get_by_filter(context,
                                                        vnfdId=vnfd_id)
            if insts:
                return

        # prevent raising exception since this method is not a part of VNFM.
        try:
            vnf_pkg.usage_state = state
            vnf_pkg.save()
        except Exception as ex:
            LOG.error("Update vnfPackage %s to %s failed: %s",
                      pkg_vnfd.package_uuid, state, ex)

    def recv_inst_create_notification(self, context, inst):
        # update vnfPackage usageState to IN_USE
        self._update_vnf_pkg_usage_state(context, inst.vnfdId,
            fields.PackageUsageStateType.IN_USE)

    def recv_inst_delete_notification(self, context, inst):
        # update vnfPackage usageState to NOT_IN_USE
        self._update_vnf_pkg_usage_state(context, inst.vnfdId,
            fields.PackageUsageStateType.NOT_IN_USE)

    def _glance_delete_images(self, vim_info, inst_id):
        # prevent raising exception since this method is not a part of VNFM.
        try:
            glance_client = glance_utils.GlanceClient(vim_info)
            images = glance_client.list_images(tag=inst_id)
        except Exception:
            LOG.error("Get glance images for vnfInstance %s failed.", inst_id)
            return

        for image in images:
            try:
                glance_client.delete_image(image.id)
                LOG.debug("image deleted name: %s id: %s",
                        image.name, image.id)
            except Exception:
                LOG.error("image delete %s failed.", image.id)

    def recv_lcmocc_notification(self, context, lcmocc, inst):
        if lcmocc.operation == v2_fields.LcmOperationType.TERMINATE:
            if (lcmocc.operationState ==
                    v2_fields.LcmOperationStateType.PROCESSING):
                # register vim_info of vnf instance so that
                # it is used later to delete glance image.
                vim_info = inst_utils.select_vim_info(inst.vimConnectionInfo)
                self.inst_vim_info[inst.id] = vim_info
            elif (lcmocc.operationState ==
                    v2_fields.LcmOperationStateType.FAILED_TEMP):
                self.inst_vim_info.pop(inst.id, None)
            elif (lcmocc.operationState ==
                    v2_fields.LcmOperationStateType.COMPLETED):
                vim_info = self.inst_vim_info.pop(inst.id, None)
                if vim_info is None:
                    # never happen. just for code consistency.
                    return
                self._glance_delete_images(vim_info, inst.id)
        elif lcmocc.operation == v2_fields.LcmOperationType.MODIFY_INFO:
            if (lcmocc.operationState ==
                    v2_fields.LcmOperationStateType.PROCESSING):
                # register vnfdId of vnf instance so that
                # it is used later to check vnfdId change.
                self.inst_vnfd_id[inst.id] = inst.vnfdId
            elif (lcmocc.operationState ==
                    v2_fields.LcmOperationStateType.FAILED_TEMP):
                self.inst_vnfd_id.pop(inst.id, None)
            elif (lcmocc.operationState ==
                    v2_fields.LcmOperationStateType.COMPLETED):
                vnfd_id = self.inst_vnfd_id.pop(inst.id, None)
                if vnfd_id is None:
                    # never happen. just for code consistency.
                    return
                if vnfd_id != inst.vnfdId:
                    # vnfdId is changed. change usage_state of vnf package
                    self._update_vnf_pkg_usage_state(context, vnfd_id,
                        fields.PackageUsageStateType.NOT_IN_USE)
                    self._update_vnf_pkg_usage_state(context, inst.vnfdId,
                        fields.PackageUsageStateType.IN_USE)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
from datetime import datetime

from oslo_db.exception import DBDuplicateEntry
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy import schema

from tacker.common import exceptions
from tacker.db.common_services import common_services_db_plugin
from tacker.db import db_base
from tacker.db import model_base
from tacker.db import models_v1
from tacker.db import types
from tacker.extensions import nfvo
from tacker.extensions.nfvo_plugins import network_service
from tacker.plugins.common import constants

LOG = logging.getLogger(__name__)
_ACTIVE_UPDATE = (constants.ACTIVE, constants.PENDING_UPDATE)
_ACTIVE_UPDATE_ERROR_DEAD = (
    constants.PENDING_CREATE, constants.ACTIVE, constants.PENDING_UPDATE,
    constants.PENDING_DELETE, constants.ERROR, constants.DEAD)
CREATE_STATES = (constants.PENDING_CREATE, constants.DEAD)


###########################################################################
# db tables

class NSD(model_base.BASE, models_v1.HasId, models_v1.HasTenant,
        models_v1.Audit):
    """Represents NSD to create NS."""

    __tablename__ = 'nsd'
    # Descriptive name
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text)
    vnfds = sa.Column(types.Json, nullable=True)

    # Nsd template source - onboarded
    template_source = sa.Column(sa.String(255), server_default='onboarded')

    # (key, value) pair to spin up
    attributes = orm.relationship('NSDAttribute',
                                  backref='nsd')

    __table_args__ = (
        schema.UniqueConstraint(
            "tenant_id",
            "name",
            name="uniq_nsd0tenant_id0name"),
    )


class NSDAttribute(model_base.BASE, models_v1.HasId):
    """Represents attributes necessary for creation of ns in (key, value) pair

    """

    __tablename__ = 'nsd_attribute'
    nsd_id = sa.Column(types.Uuid, sa.ForeignKey('nsd.id'),
            nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.TEXT(65535), nullable=True)


class NS(model_base.BASE, models_v1.HasId, models_v1.HasTenant,
        models_v1.Audit):
    """Represents network services that deploys services.

    """

    __tablename__ = 'ns'
    nsd_id = sa.Column(types.Uuid, sa.ForeignKey('nsd.id'))
    nsd = orm.relationship('NSD')

    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text, nullable=True)

    # Dict of VNF details that network service launches
    vnf_ids = sa.Column(sa.TEXT(65535), nullable=True)

    # VNFFG ids
    vnffg_ids = sa.Column(sa.TEXT(65535), nullable=True)

    # Dict of mgmt ip addresses that network service launches
    mgmt_ip_addresses = sa.Column(sa.TEXT(65535), nullable=True)

    status = sa.Column(sa.String(64), nullable=False)
    vim_id = sa.Column(types.Uuid, sa.ForeignKey('vims.id'), nullable=False)
    error_reason = sa.Column(sa.Text, nullable=True)

    __table_args__ = (
        schema.UniqueConstraint(
            "tenant_id",
            "name",
            name="uniq_ns0tenant_id0name"),
    )


class NSPluginDb(network_service.NSPluginBase, db_base.CommonDbMixin):

    def __init__(self):
        super(NSPluginDb, self).__init__()
        self._cos_db_plg = common_services_db_plugin.CommonServicesPluginDb()

    def _get_resource(self, context, model, id):
        try:
            return self._get_by_id(context, model, id)
        except orm_exc.NoResultFound:
            if issubclass(model, NSD):
                raise network_service.NSDNotFound(nsd_id=id)
            if issubclass(model, NS):
                raise network_service.NSNotFound(ns_id=id)
            else:
                raise

    def _get_ns_db(self, context, ns_id, current_statuses):
        try:
            ns_db = (
                self._model_query(context, NS).
                filter(NS.id == ns_id).
                filter(NS.status.in_(current_statuses)).
                with_for_update().one())
        except orm_exc.NoResultFound:
            raise network_service.NSNotFound(ns_id=ns_id)
        return ns_db

    def _update_ns_db(self, ns_db, new_status):
        ns_db.update({'status': new_status})
        return ns_db

    def _make_attributes_dict(self, attributes_db):
        return dict((attr.key, attr.value) for attr in attributes_db)

    def _make_nsd_dict(self, nsd, fields=None):
        res = {
            'attributes': self._make_attributes_dict(nsd['attributes']),
        }
        key_list = ('id', 'tenant_id', 'name', 'description',
                    'created_at', 'updated_at', 'vnfds', 'template_source')
        res.update((key, nsd[key]) for key in key_list)
        return self._fields(res, fields)

    def _make_dev_attrs_dict(self, dev_attrs_db):
        return dict((arg.key, arg.value) for arg in dev_attrs_db)

    def _make_ns_dict(self, ns_db, fields=None):
        LOG.debug('ns_db %s', ns_db)
        res = {}
        key_list = ('id', 'tenant_id', 'nsd_id', 'name', 'description',
                    'vnf_ids', 'vnffg_ids', 'status', 'mgmt_ip_addresses',
                    'error_reason', 'vim_id', 'created_at', 'updated_at')
        res.update((key, ns_db[key]) for key in key_list)
        return self._fields(res, fields)

    def create_nsd(self, context, nsd):
        vnfds = nsd['vnfds']
        nsd = nsd['nsd']
        LOG.debug('nsd %s', nsd)
        tenant_id = self._get_tenant_id_for_create(context, nsd)
        template_source = nsd.get('template_source')

        try:
            with context.session.begin(subtransactions=True):
                nsd_id = uuidutils.generate_uuid()
                nsd_db = NSD(
                    id=nsd_id,
                    tenant_id=tenant_id,
                    name=nsd.get('name'),
                    vnfds=vnfds,
                    description=nsd.get('description'),
                    deleted_at=datetime.min,
                    template_source=template_source)
                context.session.add(nsd_db)
                for (key, value) in nsd.get('attributes', {}).items():
                    attribute_db = NSDAttribute(
                        id=uuidutils.generate_uuid(),
                        nsd_id=nsd_id,
                        key=key,
                        value=value)
                    context.session.add(attribute_db)
        except DBDuplicateEntry as e:
            raise exceptions.DuplicateEntity(
                _type="nsd",
                entry=e.columns)
        LOG.debug('nsd_db %(nsd_db)s %(attributes)s ',
                  {'nsd_db': nsd_db,
                   'attributes': nsd_db.attributes})
        nsd_dict = self._make_nsd_dict(nsd_db)
        LOG.debug('nsd_dict %s', nsd_dict)
        self._cos_db_plg.create_event(
            context, res_id=nsd_dict['id'],
            res_type=constants.RES_TYPE_NSD,
            res_state=constants.RES_EVT_ONBOARDED,
            evt_type=constants.RES_EVT_CREATE,
            tstamp=nsd_dict[constants.RES_EVT_CREATED_FLD])
        return nsd_dict

    def delete_nsd(self, context, nsd_id, soft_delete=True):
        with context.session.begin(subtransactions=True):
            nss_db = context.session.query(NS).filter_by(
                nsd_id=nsd_id).first()
            if nss_db is not None and nss_db.deleted_at is None:
                raise nfvo.NSDInUse(nsd_id=nsd_id)

            nsd_db = self._get_resource(context, NSD, nsd_id)
            if soft_delete:
                nsd_db.update({'deleted_at': timeutils.utcnow()})
                self._cos_db_plg.create_event(
                    context, res_id=nsd_db['id'],
                    res_type=constants.RES_TYPE_NSD,
                    res_state=constants.RES_EVT_NA_STATE,
                    evt_type=constants.RES_EVT_DELETE,
                    tstamp=nsd_db[constants.RES_EVT_DELETED_FLD])
            else:
                context.session.query(NSDAttribute).filter_by(
                    nsd_id=nsd_id).delete()
                context.session.delete(nsd_db)

    def get_nsd(self, context, nsd_id, fields=None):
        nsd_db = self._get_resource(context, NSD, nsd_id)
        return self._make_nsd_dict(nsd_db)

    def get_nsds(self, context, filters, fields=None):
        if ('template_source' in filters) and \
                (filters['template_source'][0] == 'all'):
            filters.pop('template_source')
        return self._get_collection(context, NSD,
                                    self._make_nsd_dict,
                                    filters=filters, fields=fields)

    # reference implementation. needs to be overrided by subclass
    def create_ns(self, context, ns):
        LOG.debug('ns %s', ns)
        ns = ns['ns']
        tenant_id = self._get_tenant_id_for_create(context, ns)
        nsd_id = ns['nsd_id']
        vim_id = ns['vim_id']
        name = ns.get('name')
        ns_id = ns['ns_id']
        description = None
        if 'description' in ns:
            description = ns.get('description')
        try:
            with context.session.begin(subtransactions=True):
                if description is None:
                    nsd_db = self._get_resource(context, NSD,
                                                nsd_id)
                    description = nsd_db.description
                ns_db = NS(id=ns_id,
                           tenant_id=tenant_id,
                           name=name,
                           description=description,
                           vnf_ids=None,
                           vnffg_ids=None,
                           status=constants.PENDING_CREATE,
                           mgmt_ip_addresses=None,
                           nsd_id=nsd_id,
                           vim_id=vim_id,
                           error_reason=None,
                           deleted_at=datetime.min)
                context.session.add(ns_db)
        except DBDuplicateEntry as e:
            raise exceptions.DuplicateEntity(
                _type="ns",
                entry=e.columns)
        evt_details = "NS UUID assigned."
        self._cos_db_plg.create_event(
            context, res_id=ns_id,
            res_type=constants.RES_TYPE_NS,
            res_state=constants.PENDING_CREATE,
            evt_type=constants.RES_EVT_CREATE,
            tstamp=ns_db[constants.RES_EVT_CREATED_FLD],
            details=evt_details)
        return self._make_ns_dict(ns_db)

    def create_ns_post(self, context, ns_id, mistral_obj,
            vnfd_dict, vnffgd_templates, error_reason):
        LOG.debug('ns ID %s', ns_id)
        output = ast.literal_eval(mistral_obj.output)
        mgmt_ip_addresses = dict()
        vnf_ids = dict()
        vnffg_ids = dict()
        if len(output) > 0:
            for vnfd_name, vnfd_val in vnfd_dict.items():
                for instance in vnfd_val['instances']:
                    if 'mgmt_ip_address_' + instance in output:
                        mgmt_ip_addresses[instance] = ast.literal_eval(
                            output['mgmt_ip_address_' + instance].strip())
                        vnf_ids[instance] = output['vnf_id_' + instance]
            vnf_ids = str(vnf_ids)
            mgmt_ip_addresses = str(mgmt_ip_addresses)
            if vnffgd_templates:
                for vnffg_name in vnffgd_templates:
                    vnffg_output = 'vnffg_id_%s' % vnffg_name
                    vnffg_ids[vnffg_name] = output[vnffg_output]
            vnffg_ids = str(vnffg_ids)

        if not vnf_ids:
            vnf_ids = None
        if not mgmt_ip_addresses:
            mgmt_ip_addresses = None
        if not vnffg_ids:
            vnffg_ids = None
        status = constants.ACTIVE if mistral_obj.state == 'SUCCESS' \
            else constants.ERROR

        with context.session.begin(subtransactions=True):
            ns_db = self._get_resource(context, NS, ns_id)
            ns_db.update({'vnf_ids': vnf_ids})
            ns_db.update({'vnffg_ids': vnffg_ids})
            ns_db.update({'mgmt_ip_addresses': mgmt_ip_addresses})
            ns_db.update({'status': status})
            ns_db.update({'error_reason': error_reason})
            ns_db.update({'updated_at': timeutils.utcnow()})
            ns_dict = self._make_ns_dict(ns_db)

        self._cos_db_plg.create_event(
            context, res_id=ns_dict['id'],
            res_type=constants.RES_TYPE_NS,
            res_state=constants.RES_EVT_NA_STATE,
            evt_type=constants.RES_EVT_UPDATE,
            tstamp=ns_dict[constants.RES_EVT_UPDATED_FLD])
        return ns_dict

    # reference implementation. needs to be overrided by subclass
    def delete_ns_pre(self, context, ns_id, force_delete=False):
        with context.session.begin(subtransactions=True):
            ns_db = self._get_ns_db(
                context, ns_id, _ACTIVE_UPDATE_ERROR_DEAD)
            if not force_delete:
                if (ns_db is not None and ns_db.status in
                        [constants.PENDING_DELETE,
                         constants.PENDING_CREATE,
                         constants.PENDING_UPDATE]):
                    raise network_service.NSInUse(ns_id=ns_id)
            ns_db = self._update_ns_db(ns_db, constants.PENDING_DELETE)
        deleted_ns_db = self._make_ns_dict(ns_db)
        self._cos_db_plg.create_event(
            context, res_id=ns_id,
            res_type=constants.RES_TYPE_NS,
            res_state=deleted_ns_db['status'],
            evt_type=constants.RES_EVT_DELETE,
            tstamp=timeutils.utcnow(), details="NS delete initiated")
        return deleted_ns_db

    def delete_ns_post(self, context, ns_id, mistral_obj,
                       error_reason, soft_delete=True, force_delete=False):
        ns = self.get_ns(context, ns_id)
        nsd_id = ns.get('nsd_id')
        with context.session.begin(subtransactions=True):
            if force_delete:
                query = (
                    self._model_query(context, NS).
                    filter(NS.id == ns_id))
            else:
                query = (
                    self._model_query(context, NS).
                    filter(NS.id == ns_id).
                    filter(NS.status == constants.PENDING_DELETE))
            if not force_delete and (mistral_obj
                                     and mistral_obj.state == 'ERROR'):
                query.update({'status': constants.ERROR})
                self._cos_db_plg.create_event(
                    context, res_id=ns_id,
                    res_type=constants.RES_TYPE_NS,
                    res_state=constants.ERROR,
                    evt_type=constants.RES_EVT_DELETE,
                    tstamp=timeutils.utcnow(),
                    details="NS Delete ERROR")
            else:
                if soft_delete:
                    deleted_time_stamp = timeutils.utcnow()
                    query.update({'deleted_at': deleted_time_stamp})
                    self._cos_db_plg.create_event(
                        context, res_id=ns_id,
                        res_type=constants.RES_TYPE_NS,
                        res_state=constants.PENDING_DELETE,
                        evt_type=constants.RES_EVT_DELETE,
                        tstamp=deleted_time_stamp,
                        details="ns Delete Complete")
                else:
                    query.delete()
            try:
                template_db = self._get_resource(context, NSD, nsd_id)
                if template_db.get('template_source') == 'inline':
                    self.delete_nsd(context, nsd_id)
            except orm_exc.NoResultFound:
                pass

    def get_ns(self, context, ns_id, fields=None):
        ns_db = self._get_resource(context, NS, ns_id)
        return self._make_ns_dict(ns_db)

    def get_nss(self, context, filters=None, fields=None):
        return self._get_collection(context, NS,
                                    self._make_ns_dict,
                                    filters=filters, fields=fields)



# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add tenant_id to lcm_subscriptions and lcm_op_occs

Revision ID: d6ae359ab0d6
Revises: 3ff50553e9d3
Create Date: 2022-01-06 13:35:53.868106

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd6ae359ab0d6'
down_revision = '3ff50553e9d3'


def upgrade(active_plugins=None, options=None):
    op.add_column('vnf_lcm_subscriptions',
                  sa.Column('tenant_id', sa.String(length=64),
                  nullable=False))

    op.add_column('vnf_lcm_op_occs',
                  sa.Column('tenant_id', sa.String(length=64),
                  nullable=False))

    op.add_column('vnf_resources',
                  sa.Column('tenant_id', sa.String(length=64),
                  nullable=False))



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from urllib import parse

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import strutils

from tacker._i18n import _
from tacker.vnfm.monitor_drivers.token import Token
from tacker import wsgi
# check alarm url with db --> move to plugin


LOG = logging.getLogger(__name__)

OPTS = [
    cfg.StrOpt('username', default='admin',
        help=_('User name for alarm monitoring')),
    cfg.StrOpt('password', default='devstack',
        help=_('Password for alarm monitoring')),
    cfg.StrOpt('project_name', default='admin',
        help=_('Project name for alarm monitoring')),
    cfg.StrOpt('user_domain_name', default='default',
        help=_('User domain name for alarm monitoring')),
    cfg.StrOpt('project_domain_name', default='default',
        help=_('Project domain name for alarm monitoring')),
]

cfg.CONF.register_opts(OPTS, 'alarm_auth')


def config_opts():
    return [('alarm_auth', OPTS)]


class AlarmReceiver(wsgi.Middleware):
    def process_request(self, req):
        LOG.debug('Process request: %s', strutils.mask_password(req))
        if req.method != 'POST':
            return
        url = req.url
        if not self.handle_url(url):
            return
        prefix, info, params = self.handle_url(req.url)
        resource = 'trigger' if info[4] != 'maintenance' else 'maintenance'
        redirect = resource + 's'
        auth = cfg.CONF.keystone_authtoken
        alarm_auth = cfg.CONF.alarm_auth
        token = Token(username=alarm_auth.username,
                      password=alarm_auth.password,
                      project_name=alarm_auth.project_name,
                      auth_url=auth.auth_url + '/v3',
                      user_domain_name=alarm_auth.user_domain_name,
                      project_domain_name=alarm_auth.project_domain_name)

        token_identity = token.create_token()
        req.headers['X_AUTH_TOKEN'] = token_identity
        # Change the body request
        if req.body:
            body_dict = dict()
            body_dict[resource] = {}
            body_dict[resource].setdefault('params', {})
            # Update params in the body request
            body_info = jsonutils.loads(req.body)
            body_dict[resource]['params']['credential'] = info[6]
            if resource == 'maintenance':
                body_info.update({
                    'body': self._handle_maintenance_body(body_info)})
                del body_info['reason_data']
            else:
                # Update policy and action
                body_dict[resource]['policy_name'] = info[4]
                body_dict[resource]['action_name'] = info[5]
            body_dict[resource]['params']['data'] = body_info
            req.body = jsonutils.dump_as_bytes(body_dict)
            LOG.debug('Body alarm: %s', req.body)
        # Need to change url because of mandatory
        req.environ['PATH_INFO'] = prefix + redirect
        req.environ['QUERY_STRING'] = ''
        LOG.debug('alarm url in receiver: %s', req.url)

    def handle_url(self, url):
        # alarm_url = 'http://host:port/v1.0/vnfs/vnf-uuid/mon-policy-name/action-name/8ef785' # noqa
        parts = parse.urlparse(url)
        p = parts.path.split('/')
        if len(p) != 7:
            return None

        if any((p[0] != '', p[2] != 'vnfs')):
            return None
        # decode action name: respawn%25log
        p[5] = parse.unquote(p[5])
        qs = parse.parse_qs(parts.query)
        params = dict((k, v[0]) for k, v in qs.items())
        prefix_url = '/%(collec)s/%(vnf_uuid)s/' % {'collec': p[2],
                                                    'vnf_uuid': p[3]}
        return prefix_url, p, params

    def _handle_maintenance_body(self, body_info):
        body = {}
        traits_list = body_info['reason_data']['event']['traits']
        if type(traits_list) is not list:
            return
        for key, t_type, val in traits_list:
            if t_type == 1 and val and (val[0] == '[' or val[0] == '{'):
                body[key] = eval(val)
            else:
                body[key] = val
        return body



# Copyright (c) 2014-2018 China Mobile (SuZhou) Software Technology Co.,Ltd.
# All Rights Reserved
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.plugins.common import constants
from tacker.vnfm.policy_actions.autoscaling import autoscaling \
    as policy_actions_autoscaling


class TestVNFActionAutoscaling(testtools.TestCase):

    def setUp(self):
        super(TestVNFActionAutoscaling, self).setUp()
        self.context = context.get_admin_context()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()

    def test_execute_action(self):
        action_autoscaling = policy_actions_autoscaling.VNFActionAutoscaling()
        vnf_dict = {
            'id': '00000000-0000-0000-0000-000000000001',
            'status': 'fake-status'}
        plugin = mock.Mock()
        plugin.create_vnf_scale.return_value = None
        action_autoscaling.execute_action(plugin, self.context, vnf_dict, None)
        self._cos_db_plugin.create_event.assert_called_once_with(
            self.context, res_id=vnf_dict['id'],
            res_state=vnf_dict['status'],
            res_type=constants.RES_TYPE_VNF,
            evt_type=constants.RES_EVT_MONITOR,
            tstamp=mock.ANY, details="ActionAutoscalingHeat invoked")
        plugin.create_vnf_scale.assert_called_once_with(
            self.context, vnf_dict['id'], None)
